package INSTALL::apache;

use INSTALL::os;
use INSTALL::query;
use INSTALL::caspeng;
use INSTALL::common;

my $debug = 0;
my $debug_output = ($debug ? "" : " > /dev/null 2>&1");
my $CONST_Package = "INSTALL::apache";

my $pGlobals = \%INSTALL::global::variables;
my $pParameters = \%INSTALL::parameter::variables;
my $pOS = \%INSTALL::os::Details;

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Internal functions
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#------------------------------------------------------------------------------
# All of the error reporting mechanisms take of a list of strings, and format
# them such that they fall on different lines.
#------------------------------------------------------------------------------
sub error {
  my $result = INSTALL::common::generate_error(@_);
  print STDERR "Package location: $CONST_Package\n";
  return($result);
}

sub trace
{
  if ($debug)
  {
    my $result = INSTALL::common::generate_report("trace: ",@_);
    print STDERR "Package location: $CONST_Package\n";
    return($result);
  }
}

sub note
{
  my $result = INSTALL::common::generate_note(@_);
  return($result);
}

sub report 
{
  my $result = INSTALL::common::generate_report("",@_);
  return($result);
}

sub query_boolean_default
{
  ($#_ == 1) || die "Invalid number of arguments.";
  my ($query_string,$default) = @_;
  my @options;
  
  if ($default) {
    @options = ("default=y","n");
  }
  else {
    @options = ("y","default=n");
  }
  
  return(INSTALL::conditional::boolean_value
	 (INSTALL::query::user($query_string,
			       "terminal_mark=?",
			       @options)));
}


#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# External functions
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#------------------------------------------------------------------------------
# This method attempts to find an apache server located in the directory 
# specified.  It does this by checking for an httpd.conf file in the conf 
# subdirectory under this directory.  If that file is found, it is returned,
# along with the type (ie. apache).
#------------------------------------------------------------------------------
sub locate_server
{
  ($#_ == 0) || die "Invalid number of arguments.";
  
  my ($server_path) = @_;

  if (-d "$server_path/conf" && -f "$server_path/conf/httpd.conf") {
    return("$server_path/conf/httpd.conf","apache");
  }
  elsif (-f "$server_path/httpd.conf") {
    return("$server_path/httpd.conf","apache");
  }
  else {
    # Make sure we match a subdirectory with the name apache in it, for 
    # completeness.
    if (opendir(ROOT,"$server_path")) {
      my @subfiles = readdir(ROOT);
      close(ROOT);

      foreach(@subfiles) {
	if ((($_ ne ".") && ($_ ne "..")) &&
	    ((-d "$server_path/$_") && /apache/i))
	{
	  my @result = locate_server("$server_path/$_");
	  if ($#result == 1) {
	    return(@result);
	  }
	}
      }
    }

    return(undef);
  }
}


#------------------------------------------------------------------------------
# This method merely discovers as much as it can about the server with the 
# information pointed to by $pHash, storing this information in the $pHash.
#------------------------------------------------------------------------------
# pHash is guaranteed to contain:
#   conf = <configuration file path>
#   type = apache
sub cache_server
{
  ($#_ == 0) || die "Invalid number of arguments to cache_server.";

  my ($pServerStats) = @_;
  my $autostart_enabled;

  ## Preemptively discover the start / stop scripts, if possible.
  for($pOS->{package_formats})
  {
     /RPM/i &&
       do {
	 for($pOS->{distribution_flavor}) {
	   /RedHat/i &&
	     do {
	       if ($pServerStats->{conf} eq "/etc/httpd/conf/httpd.conf") {
		 $pServerStats->{start_script} = "/etc/rc.d/init.d/httpd start";
		 $pServerStats->{stop_script} =  "/etc/rc.d/init.d/httpd stop";
		 $autostart_enabled = 1;
	       }
	       last;
	     };
	   
	   /SuSE/i &&
	     do {
	       if ($pServerStats->{conf} eq "/etc/httpd/httpd.conf") {
		 $pServerStats->{start_script} = "/sbin/init.d/apache start";
		 $pServerStats->{stop_script} = "/sbin/init.d/apache stop";
		 $autostart_enabled = 1;
	       }
	       last;
	     };
	 };

	 # Else, ignore.
	 last;
       };

     /DEB/i &&
       do {
	 # Nothing at the moment, but Debian may soon be coming.
	 last;
       };

     /TAR/i &&
       do {
	 for($pOS->{distribution_flavor}) {
	   /Slackware/i &&
	     do {
	       if ($pServer->{conf} eq "/var/lib/apache/conf/httpd.conf") {
		 $pServerStats->{start_script} = "/var/lib/apache/sbin/apachectl start";
		 $pServerStats->{stop_script} = "/var/lib/apache/sbin/apachectl stop";
		 $autostart_enabled = 1;
	       }
	       last;
	     };
	 }
	 
	 last;
       };
   }

  unless(open(CONFIG, "<$pServerStats->{conf}")) {
    return(error("Could not open $pServerStats->{conf} - cache_server."));
  }

  while(<CONFIG>) {
    if (/^\s*ServerRoot\s+(\S+)/) {
      $pServerStats->{root} = $1;
    }
    elsif (/^\s*Port\s+([0-9]+)/) {
      $pServerStats->{port} = $1;
    }
  }

  close(CONFIG);

  # Remove any "" encapsulation.
  foreach(keys %{$pServerStats}) {
    $pServerStats->{$_} =~ s/\"([^\"]*)\"/$1/;
  }

  if (not exists $pServerStats->{binary}) {
    if (exists $pParameters->{webserver_binary}) {
      if (-x $pParameters->{webserver_binary}) {
	my $valid_apache = `$pParameters->{webserver_binary} -v`;
	
	if ($valid_apache !~ /apache/i) {
	  note("The specified webserver_binary=$pParameters->{webserver_binary} option",
	       "specifies a binary which does not appear to be an Apache binary. This",
	       "was detected by running '$pParameters->{webserver_binary} -v' and not",
	       "receiving a result containing the case-insensitive word 'Apache'.  As",
	       "a result while the install will attempt to continue, the binary",
	       "assumed to correspond with the webserver configuration file (specified",
	       "below) may not be correct:",
	       "  $pServerStats->{conf}");
	}
	else {
	  $pServerStats->{binary} = $pParameters->{webserver_binary};
	}
      }
      else {
	note("The specified webserver_binary=$pParameters->{webserver_binary} option",
	     "is invalid.  It specifies a file that either does not exist or is not",
	     "set executable.  While the install will attempt to continue, the binary",
	     "assumed to correspond with the webserver configuration file (specified",
	     "below) may not be correct:",
	     "  $pServerStats->{conf}");
      }
    }
    
    foreach("sbin","bin","src") {
      # Check for secure server first.
      my $subdir = $_;
      
      foreach("httpsd","httpd") {
	if (-r "$pServerStats->{root}/$subdir/$_") {
	  $pServerStats->{binary} = "$pServerStats->{root}/$subdir/$_";
	  last;
	}
      }
    }
  }

  if (not exists $pServerStats->{binary}) {
    # Check for secure server first.

    foreach("httpsd","httpd") {
      # Begin the guessing game ...
      if (-r "/usr/sbin/$_") {
	$pServerStats->{binary} = "/usr/sbin/$_";
	last;
      }
      elsif (-r "/usr/bin/$_") {
	$pServerStats->{binary} = "/usr/bin/$_";
	last;
      }
      elsif (-r "/var/lib/apache/sbin/$_") {
	$pServerStats->{binary} = "/var/lib/apache/sbin/$_";
	last;
      }
    }

    if (not exists $pServerStats->{binary}) {
      return(error("Unable to obtain several key values for the server with config file:",
		   "  $pServerStats->{conf}"));
    }
  }

  my $version_string = `$pServerStats->{binary} -v`;
  my $module_string = `$pServerStats->{binary} -l`;
  my $define_string = `$pServerStats->{binary} -V`;
  
  if ($version_string =~ m@Apache/(\d+)\.(\d+)\.(\d+)@s) {
    $pServerStats->{version} = "$1.$2.$3";
    $pServerStats->{release} = $1;
    $pServerStats->{major} = $2;
    $pServerStats->{minor} = $3;
  }
  else {
    return(error("Unable to obtain version values for the server with config file:",
		 "  $pServerStats->{conf}"));
  }
  
  
  if ($module_string =~ /mod_so/) {
    $pServerStats->{dso} = "yes";
  }
  else {
    $pServerStats->{dso} = "no";
  }

  $pServerStats->{api} = "";
  ########## FIXME:
  ########## There is no real way of detecting FrontPage inside of an apache binary
  ########## (due to slipshot workmanship by Ready to Run Software).  They broke
  ########## binary compatibility (even though they didn't need to), and as a result
  ########## we must hardcode in when and when not to enable the FrontPage module.
  ########## This is only currently needed on Mandrake.
  if (($pOS->{distribution} =~ /Mandrake/i) && 
      ($pOS->{distribution_version} =~ /^([789]|[1-9][0-9]+)/))
  {
    # Frontpage support is enabled on Mandrake 7.0+.
    $pServerStats->{api} .= "frontpage-";
  }
  
  if ($define_string =~ /EAPI/) {
    $pServerStats->{api} .= "eapi";
  }
  else {
    $pServerStats->{api} .= "standard";
  }

  # We should refresh the start script, every time we run cache server.  The only
  # exception is if the start_script was set because of a well known location
  # for that start script (ie. a system dependent start script was encountered).
  if ((not $autostart_enabled) || (not exists $pServerStats->{start_script})) {
    # Let's see if we can make a last ditch effort at generating a start / stop
    # script.

    if (-x "$pServerStats->{root}/apachectl") {
      $pServerStats->{start_script} = "$pServerStats->{root}/apachectl start";
      $pServerStats->{stop_script} = "$pServerStats->{root}/apachectl stop";
    }
    elsif (-x $pServerStats->{binary}) {
      $pServerStats->{start_script} = 
	("$pGlobals->{asphome}/INSTALL/apachectl \"binary=$pServerStats->{binary}\" ".
	 "\"conf=$pServerStats->{conf}\" start");

      $pServerStats->{stop_script} = 
	("$pGlobals->{asphome}/INSTALL/apachectl \"binary=$pServerStats->{binary}\" ".
	 "\"conf=$pServerStats->{conf}\" stop");
    }
  }

  # Detect secure server ability.
  if ($pServerStats->{binary} =~ m@httpsd@) {
    $pServerStats->{https} = "yes";
  }
  else {
    $pServerStats->{https} = "no";
  }

  return(1);
}

#------------------------------------------------------------------------------
# This method will ask all of the appropriate questions, to properly configure
# an apache server.  This method will note, however, do the actual 
# installation.  The name is a bit misleading, since it in truth only gathers
# the information necessary for configuring the server.
#------------------------------------------------------------------------------
sub configure_unlisted
{
  sub confirm_existence {
    ($#_ == 0) || die "Invalid number of arguments.";
    
    my ($file) = @_;
    
    if ($file eq "none") {
      report("");
      my $answer = INSTALL::query::user
	("Would you like to cancel the set-up","terminal_mark=?","y","default=n");
      report("");

      if ($answer eq "y") {
	return(1);
      }
      else {
	# Cause a requery.
	return(undef);
      }
    }
    elsif (not -f $file) {
      report("Invalid: The specified file does not exist.");
      report("");
    }

    return(-f $file);
  }
  
  sub confirm_binary_existence {
    ($#_ == 0) || die "Invalid number of arguments.";
    
    my ($file) = @_;
    
    if ($file eq "none") {
      report("");
      my $answer = INSTALL::query::user
	("Would you like to cancel the set-up","terminal_mark=?","y","default=n");
      report("");

      if ($answer eq "y") {
	return(1);
      }
      else {
	# Cause a requery.
	return(undef);
      }
    }
    elsif (not -x $file) {
      report("Invalid: The specified file does not exist or is not executable.");
      report("");
    }
    
    return(-x $file);
  }

  INSTALL::query::user_menu
    ("CONFIGURE AN UNLISTED APACHE SERVER",
     "Because of the way Apache functions, very few of the configuration options ".
     "can be answered outright, automatically.  However, each question ".
     "where we have a guess available, will have a default value listed next to ".
     "the question.  If you wish to terminate the install, merely type none for ".
     "the listed option."
    );


  my $server_conf = INSTALL::query::user_function
    ("Input the path and file name of the Apache configuration file\n",
     *confirm_existence, "default=none", "invalid_choice=");

  if ($server_conf eq "none") {
    return(undef);
  }

  my %hash = 
    (conf => $server_conf,
     type => apache);

  my $pServerStats = \%hash;

  cache_server($pServerStats);

  print "\n\n";

  my $server_binary = INSTALL::query::user_function
    ("Input the path and file name of the Apache binary\n",
     *confirm_binary_existence,
     "default=$pServerStats->{binary}",
     "invalid_choice=");

  if ($server_binary eq "none") {
    return(undef);
  }

  if ((not exists $pServerStats->{binary}) || 
      ($server_binary ne $pServerStats->{binary})) 
  {
    # Reconfigure, if things changed due to a differing binary.
    $hash{binary} = $server_binary;
    cache_server(\%hash);
  }
  
  return(%hash);
}

##################################################
# Start / Stop Apache rpm on (X) distribution.
##################################################

sub restart_apache
{
  my (%server_info) = @_;

  # Handle both input from .installed_db and .webserver_cache
  foreach(keys %server_info) {
    my $key = $_;

    if (/^webserver_(.*)$/) {
      $server_info{$1} = $server_info{$key};
    }
  }

  if (exists $server_info{stop_script} && exists $server_info{start_script}) {
    my $result = not system("$server_info{stop_script} $debug_output");
    INSTALL::common::usleep(500000);
    $result |= not system("$server_info{start_script} $debug_output");
    return($result);
  }
  else {
    return(0);
  }
}

sub start_apache
{
  my (%server_info) = @_;

  # Handle both input from .installed_db and .webserver_cache
  foreach(keys %server_info) {
    my $key = $_;

    if (/^webserver_(.*)$/) {
      $server_info{$1} = $server_info{$key};
    }
  }

  if (exists $server_info{start_script}) {
    my $result = not system("$server_info{start_script} $debug_output");
    return($result);
  }
  else {
    return(0);
  }
}

sub stop_apache
{
  my (%server_info) = @_;

  # Handle both input from .installed_db and .webserver_cache
  foreach(keys %server_info) {
    my $key = $_;

    if (/^webserver_(.*)$/) {
      $server_info{$1} = $server_info{$key};
    }
  }

  if (exists $server_info{stop_script}) {
    my $result = not system("$server_info{stop_script} $debug_output");
    INSTALL::common::usleep(500000);
    return($result);
  }
  else {
    return(0);
  }
}

#------------------------------------------------------------------------------
# Perform the job of apxs. If a fourth argument value (which evaluates to true),
# does not exist, the default semantics of apxs are used.  That is, the new
# module is appended to the end of the AddModule / LoadModule lists.  
# Otherwise, if it does evaluate to true, the module will be inserted at the
# head of the list (ie. prepended).
#------------------------------------------------------------------------------
sub apxs
{
  (($#_ == 1) || ($#_ == 2)) || die "Invalid number of arguments.";

  my ($conf,$entry_point,$module_name,$prepend) = @_;
  my $dupname = "$conf.pre-chiliasp-apxs";

  if (not -f $module_name) {
    return(error("Specified module '$module_name' does not exist."));
  }

  unlink($dupname);
  system("/bin/cp $conf $dupname");
  unless(open(SRC,"<$dupname") && open(HTTPD,">$conf")) 
  {
    # Back out any changes.
    close(SRC);
    close(HTTPD);

    INSTALL::caspeng::uninstall(%server_info);
    unlink("$conf");
    system("/bin/cp $conf.pre-chiliasp $conf");
    unlink("$conf.pre-chiliasp");
    
    return(error("Unable to open $conf for modification."));
  }

  my $module_root = INSTALL::common::strip_extension
    (INSTALL::common::extract_filename($module_name));

  my $quoted_module_root = quotemeta($module_root);

  foreach(<SRC>) {
    if ((not m@[/\s]$quoted_module_root@) && (not /Chili!Soft/)) {
      push @lines,$_;
    }
  }

  my @result;
  my $line_count = $#lines;

  if (grep { /^\s*ClearModuleList/ } @lines) {
    ### First find the ClearModuleList line (or actually the first 
    ### non-comment-line, before that).
    my $ClearModule_index;

    # Accomodate SuSE bug of duplicating the LoadModule / AddModule section in
    # the httpd.conf file.  We make sure we add ourself around the last 
    # ClearModuleList, rather than the first.
    for(my $i=$line_count; $i >= 0; $i--) {
      if ($lines[$i] =~ /^\s*ClearModuleList/) {
	$ClearModule_index = $i;
	last;
      }
    }

    # Search for first previous non-comment line.
    for(my $i = $ClearModule_index - 1; $i >= 0; $i--) {
      if ($lines[$i] !~ /^\s*\#/) {
	$ClearModule_index = $i;
	last;
      }
    }
    
    ### Then find the last AddModule index (or actually the first
    ### non-comment-line, after that).
    my $last_AddModule_index;
    for(my $i=$line_count; $i >= 0; $i--) {
      if ($lines[$i] =~ /^\s*AddModule/) {
	$last_AddModule_index = $i;
	last;
      }
    }

    if (not $last_AddModule_index) {
      $last_AddModule_index = $ClearModule_index + 1;
    }

    # Discover if we are in a <IfDefine> block
    my $insideIfDefine = 0;
    for(my $i = 0; $i <= $last_AddModule_index; $i++) {
      if ($lines[$i] =~ m@^\s*<IfDefine@) {
	$insideIfDefine++;
      }
      elsif ($lines[$i] =~ m@^\s*</IfDefine@) {
	$insideIfDefine--;
      }
    }

    if ($insideIfDefine > 0) {
      # We will make sure to add ourself after the appropriate closing </IfDefine>
      # (I certainly hope their httpd.conf is valid).
      for(my $i = $last_AddModule_index + 1; ($i <= $line_count); $i++) 
      {
	if ($lines[$i] =~ m@^\s*<IfDefine@) {
	  $insideIfDefine++;
	}
	elsif ($lines[$i] =~ m@^\s*</IfDefine@) {
	  $insideIfDefine--;
	}

	if ($insideIfDefine <= 0) {
	  $last_AddModule_index = $i + 1;
	  last;
	}
      }
    }

    # Find first non-comment line.
    for(my $i = $last_AddModule_index + 1; $i <= $line_count; $i++) {
      if ($lines[$i] !~ /^\s*\#/) {
	$last_AddModule_index = $i;
	last;
      }
    }
    
    # Now glue it all together.
    @lines = (
	      @lines[0 .. ($ClearModule_index - 1)],
	      "#-------------------------------------------------- Chili!Soft ASP\n",
	      "LoadModule $entry_point\t$module_name\n",
	      "#-------------------------------------------------- Chili!Soft ASP\n",
	      @lines[$ClearModule_index .. ($last_AddModule_index - 1)],
	      "#-------------------------------------------------- Chili!Soft ASP\n",
	      "AddModule $module_root.c\n",
	      "#-------------------------------------------------- Chili!Soft ASP\n",
	      @lines[$last_AddModule_index .. $line_count]
	     );
  }
  else {

    if (grep { /^\s*LoadModule/ } @lines) {
      # Put the LoadModule at the bottom of the file.
      @lines = ( @lines,
		"#-------------------------------------------------- Chili!Soft ASP\n",
		"LoadModule $entry_point\t$module_name\n",
		"#-------------------------------------------------- Chili!Soft ASP\n"
	       );
    }
    else {
      # Put the LoadModule at the top of the file.
      @lines = (
		"#-------------------------------------------------- Chili!Soft ASP\n",
		"LoadModule $entry_point\t$module_name\n",
		"#-------------------------------------------------- Chili!Soft ASP\n",
		@lines
	       );
    }
  }

  print HTTPD @lines;
  
  close(SRC);
  close(HTTPD);

  # It is now safe to wipe the slate clean.
  unlink($dupname);

  return(1);
}

#------------------------------------------------------------------------------
# This method merely attempts to take the webserver information given to it
# and tries to configure that webserver to work with a generated Chili!Soft
# ASP server.  The server is generated on the fly.
#==============================================================================
# Implementors note:
# (1) It is assumed that the caspeng module provides both a configure and 
#     uninstall script, which will handle the installation and removal of
#     a server installation.
#------------------------------------------------------------------------------
sub configure
{
  my $interactive = 
    (not INSTALL::conditional::boolean_value($pParameters->{non_interactive}));
  my $custom_install = 
    INSTALL::conditional::boolean_value($pParameters->{custom_install});

  sub detect_supported_versions {
    my @supported_versions;

    if (opendir(MODULES_DIR,$pGlobals->{module_dir})) {
      my @subdirs = readdir(MODULES_DIR);
      foreach(@subdirs) {
	if (/apache_([0-9]+\.[0-9]+\.[0-9]+)/) {
	  push @supported_versions,$1;
	}
      }

      return(@supported_versions);
    }
    else {
      error("Unable to read module directory in installation:",
	    "  $pGlobals->{module_dir}");
      exit(1);
    }
  }

  sub retrieve_server_info {
    ($#_ == 0) || die "Invalid number of arguments.";

    my ($conf) = @_;

    my $pIni = INSTALL::ini::open($pGlobals->{installed_db});
    if (!$pIni)
    {
      return(error("Could not open $pGlobals->{installed_db} - uninstall."));
    }
    
    my @section_list = INSTALL::ini::getSectionNames($pIni);
    my %server_details;

    foreach(@section_list) {
      my $pSection = INSTALL::ini::openSection($pIni,$_);
      if ($pSection->{webserver_conf} eq $conf) {
	%server_details = %{$pSection};
      }
    }

    return(%server_details);    
  }

  sub exit_parent {
    ($#_ == 0) || die "Invalid number of arguments.";

    my ($pid) = @_;

    if ($interactive) {
      INSTALL::common::signal_spinner($pid);
    }
    else {
      wait;
    }    
    return(undef);
  }

  my %server_stats = @_;
  my @supported_versions = detect_supported_versions;

  ########## Check the server_stats for a valid setup.
  if (not -d $server_stats{root}) {
    return(error("The specified webserver root '$server_stats{root}' does not exist."));
  }
  
  if (not -f $server_stats{conf}) {
    return(error("The specified webserver configuration file '$server_stats{conf}'",
		 "does not exist."));
  }
  
  if (not -x $server_stats{binary}) {
    return(error("The specified webserver executable '$server_stats{binary}' does not",
		 "exist or is not executable."));
  }

  if ($server_stats{type} !~ /apache/i) {
    return(error("Attempting to configure a '$server_stats{type}' server with the",
		 "Apache module."));
  }

  # Now that the configuration appears correct, we should recache the server 
  # with the same information, so that the DSO flag can be adjusted between
  # execution cycles.  Otherwise, the user would be required to refresh their
  # webserver list.
  my $pRecache = \%server_stats;
  
  if (cache_server($pRecache)) {
    # Override values in the server cache, by adding them to the end of the 
    # list-based insert.
    %server_stats = (%server_stats,%{$pRecache});
  }

  if (not INSTALL::conditional::boolean_value($server_stats{dso})) {
    if (not $interactive) {
      return(error("I have detected that the specified webserver does not use dynamic",
		   "modules (DSO), yet the installer requires this.  Please recompile",
		   "your webserver to support DSO.  Refer to Apache documentation on",
		   "enabling DSO support."));
    }
    else {
      INSTALL::query::user_menu
	("NON-DSO APACHE CONFIGURATION",
	 "I have detected that the specified webserver does not use dynamic modules (DSO).  ".
	 "It should be noted that to get Chili!Soft ASP functional, you have two options.  ".
	 "If you would prefer to continue on with the installation, you will be required ".
	 "to recompile your Apache webserver with the Chili!Soft ASP module (refer to the ".
	 "Chili!Soft ASP README for details).  Another option is to skip the ".
	 "installation in favor of recompiling DSO support into Apache.  While both ".
	 "options will work equally well, it is suggested that you enable DSO support, ".
	 "since it has numerous benefits over the statically linked version (refer to ".
	 "Apache documentation for details).\n \n".
	 "Note: If you opted to enable DSO support in Apache (ie. answered yes to\n".
	 "      the question below), you will need to perform the following steps,\n".
	 "      once the necessary Apache modifications have been made, to\n".
	 "      finalize your install:\n".
	 "         cd $pGlobals->{asphome}\n".
	 "         ./add_server"
	);

      my $skip_install = INSTALL::query::user
	("Would you like to skip installation to enable DSO support",
	 "default=y","n", "terminal_mark=?");
      
      if ($skip_install eq "y") {
	return(undef);
      }
    }
  }
  
  ########## If the user wants auto startup of apache and it's supported ...
  my $start_apache = 0;
  my $supported_autostart = 
    (exists $server_stats{start_script} && exists $server_stats{stop_script});
  
  if ($supported_autostart) {
    $start_apache = 1;
    
    if (exists $pParameters->{start_apache}) {
      $start_apache = INSTALL::conditional::boolean_value($pParameters->{start_apache});
    }
  }
  
  ### FIXME: We probably need a more distribution specific section of code here.
  if ($interactive) {
    if ($supported_autostart && $custom_install) {
      $start_apache = query_boolean_default
	("Would you like me to restart this Apache server for you", $start_apache);
    }
  }

  # Check version support.
  my $quoted_version = quotemeta($server_stats{version});
  
  if (not grep { /$quoted_version/ } @supported_versions) {
    error("The specified version ($version) of Apache is not currently supported.");
    note("Regardless of this fact, you may be able to build your own module that",
	 "supports this version of Apache.  Refer to Chili!Soft ASP documentation",
	 "for information on building your own module.  And, after that module has",
	 "been built, proceed by following these steps:",
	 "  (1) mkdir -p $pGlobals->{asphome}/module/apache_$version",
	 "  (2) cp <generated build dir>/mod_casp2.so $pGlobals->{asphome}/module/apache_$server_stats{version}/$server_stats{api}",
	 "  (3) Rerun the installer.",
	 "",
	 "By placing the file into the created module directory, the installer will",
	 "recognize it as a supported webserver and act accordingly.");
    report("Press enter to continue ...");
    <STDIN>;
    return(undef);
  }

  ########## Do the actual install, first by calling into caspeng::configure.
  if (INSTALL::caspeng::configure(%server_stats)) {    
    if ($interactive) {
      print STDERR ("Configuring the Apache server ... ");
    }    

    my $pid = fork;

    if ($pid) {
      ##########========== Parent
      unlink("$server_stats{conf}.pre-chiliasp");
      system("/bin/cp $server_stats{conf} $server_stats{conf}.pre-chiliasp");
      
      ########## Update the httpd.conf file to refer to the appropriate locations.
      if (INSTALL::conditional::boolean_value($server_stats{dso})) {
	if (not apxs($server_stats{conf},"casp2_module",
		     "$pGlobals->{module_dir}/apache_$server_stats{version}/$server_stats{api}/mod_casp2.so"))
	{
	  error("Unable to configure the server due to apxs failure.");

	  # Run cleanup code.
	  return(exit_parent($pid));
	}
      }
      
      my $dupname = "$server_stats{conf}.chiliasp-install";
      
      unlink($dupname);
      system("/bin/cp $server_stats{conf} $dupname");
      
      my %server_info = retrieve_server_info($server_stats{conf});
      unless(open(SRC,"<$dupname") && open(HTTPD,">$server_stats{conf}")) 
      {
	# Back out any changes.
	close(SRC);
	close(HTTPD);
	
	INSTALL::caspeng::uninstall(%server_info);
	unlink("$server_stats{conf}");
	system("/bin/cp $server_stats{conf}.pre-chiliasp $server_stats{conf}");
	unlink("$dupname");
	
	error("Unable to open $server_stats{conf} for modification.");

	# Run cleanup code.
	return(exit_parent($pid));
      }
      unlink($dupname);
      
      my $last_line;
      while (<SRC>)
      {
	$last_line = $_;
	
	s/^(MaxRequestsPerChild\s+)\d+/MaxRequestsPerChild 30000/;
	
	print HTTPD if ((! /AddHandler.*\.as[pa]|CaspLib|\#\# Chili!Soft ASP/i));
      }
      
      if (not chomp $last_line) {
	print HTTPD "\n";
      }
      
      print HTTPD "## Chili!Soft ASP installer\n";
      
      print HTTPD 
	("AddHandler chiliasp .asp\n".
	 "AddHandler chiliasp .asa\n".
	 "CaspLib $server_info{engine_dir}\n"
	);
      close HTTPD;
      close SRC;
      
      my $restarted = 0;
      
      if ($start_apache) {
	$restarted = restart_apache(%server_info);
      }
      
      exit_parent($pid);
      if ($interactive && (not $restarted)) {
	print STDERR "\n";
	note("For your webserver to function properly, you will need to restart",
	     "it.  The configuration file for this webserver is:",
	     "  $server_stats{conf}");
	report("Press enter to continue ...");
	<STDIN>;
      }
      
      return(1);
    }
    else {
      ##########========== Child
      if ($interactive) {
	INSTALL::common::spinner_wait_signal;
      }
      exit(0);
    }
  }
  else {
    return(undef);
  }
}

#------------------------------------------------------------------------------
# This method when invoked on information regarding a specific Apache server
# will remove any remnants of Chili!Soft ASP from that server, as well as
# remove any server, which points to that Apache server.
#==============================================================================
# Implementors note:
# (1) This method assumes that the caspeng module has an uninstall method which 
#     takes the information of a webserver, and properly removes itself.
# (2) The naming convention in %server_stats is different from that used during
#     the install.  As dictated in caspeng::configure, webserver_ is prepended
#     to all contained webserver information.
#------------------------------------------------------------------------------
sub uninstall
{
  my %server_stats = @_;

  if (INSTALL::caspeng::uninstall(%server_stats)) {
    unlink("$server_stats{conf}.post-chiliasp");
    system("/bin/cp $server_stats{webserver_conf} $server_stats{webserver_conf}.post-chiliasp");

    unless(open(SRC,"<$server_stats{webserver_conf}.post-chiliasp") && open(HTTPD,">$server_stats{webserver_conf}")) 
    {
      return(error("Unable to open $server_stats{webserver_conf} for modification."));
    }

    ########## Strip any Chili!Soft ASP line from the httpd.conf file.
    ########## Keep track of the PidFile directive.
    my $pidFile;
    while (<SRC>)
    {
      if (/[^\#]*PidFile\s+\"([^\"]|\\.)\"/i) {
	$pidFile = $1;
      }
      elsif (/[^\#]*PidFile\s+(\S+)/i) {
	$pidFile = $1;
      }
      print HTTPD if ((! /AddHandler.*\.as[pa]|CaspLib|Chili!Soft ASP|mod_casp2|casp2_module/i));
    }

    close HTTPD;
    close SRC;

    my $restarted = 0;
    my $interactive = (not INSTALL::conditional::boolean_value($pParameters->{non_interactive}));

    ########## Restart the webserver as appropriate.
    ### FIXME: This should be more implementation specific ... probably.
    if (-f $pidFile) {
      if (exists $server_stats{webserver_start_script} && 
	  exists $server_stats{webserver_stop_script}) 
      {
	my $answer = "y";
	
	if ($interactive) {
	  $answer = INSTALL::query::user
	    ("Would you like to restart this apache server",
	     "terminal_mark=?",
	     "default=y",
	     "n");
	}
	
	if ($answer eq "y") {
	  $restarted = restart_apache(%server_stats);
	}
      }
      
      if ($interactive && (not $restarted)) {
	note("For your webserver to function properly, you will need to restart",
	     "it.  The configuration file for this webserver is:",
	     "  $server_stats{webserver_conf}");
	report("Press enter to continue ...");
	<STDIN>;
      }
    }

    return(1);
  }
  else {
    return(undef);
  }    
}

1;
