package INSTALL::caspeng;

use INSTALL::os;
use INSTALL::common;
use INSTALL::conditional;
use INSTALL::global;
use INSTALL::parameter;
use INSTALL::services;
use INSTALL::substitute;
use INSTALL::query;
use INSTALL::reboot;
use INSTALL::ini;
use POSIX;

my $debug = 0;
my $debug_output = ($debug ? "" : "> /dev/null 2>&1");
my $disable_multiprocess = 1;

my $CONST_Package = "INSTALL::caspeng";
my $cache_file = "$pGlobals->{install_dir}/webserver-cache";

my $pGlobals = \%INSTALL::global::variables;
my $pParameters = \%INSTALL::parameter::variables;
my $pOS = \%INSTALL::os::Details;

my $CONST_StartPriority = 86;
my $CONST_StopPriority = 14;

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Internal functions
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#------------------------------------------------------------------------------
# All of the error reporting mechanisms take a list of strings, and format
# them such that they fall on different lines.
#------------------------------------------------------------------------------
sub error {
  my $result = INSTALL::common::generate_error(@_);
  print STDERR "Package location: $CONST_Package\n";
  return($result);
}

sub note
{
  my $result = INSTALL::common::generate_note(@_);
  return($result);
}

sub trace
{
  if ($debug)
  {
    my $result = INSTALL::common::generate_report("trace: ",@_);
    print STDERR "Package location: $CONST_Package\n";
    return($result);
  }
}

sub report 
{
  my $result = INSTALL::common::generate_report("",@_);
  return($result);
}

sub query_boolean_default
{
  ($#_ == 1) || die "Invalid number of arguments.";
  my ($query_string,$default) = @_;
  my @options;
  
  if ($default) {
    @options = ("default=y","n");
  }
  else {
    @options = ("y","default=n");
  }
  
  return(INSTALL::conditional::boolean_value
	 (INSTALL::query::user($query_string,
			       "terminal_mark=?",
			       @options)));
}

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# External functions
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#------------------------------------------------------------------------------
# This method returns the list of all known installed engines.
#==============================================================================
sub list_engines
{
  my $pIni = INSTALL::ini::open($pGlobals->{installed_db});
  if (!$pIni)
  {
    return(error("Could not open $pGlobals->{installed_db}."));
  }

  return(INSTALL::ini::getSectionNames($pIni));
}

#------------------------------------------------------------------------------
# The caspeng::configure method is responsible for querying / identifying the
# wants of the user and incorporating those wants into the generated engine.
#==============================================================================
# Implementors note:
# (1) The data stored into installed_db will be keyed off of the server name 
#     (ie. the directory name of the generated server) and will contain a 
#     naming scheme such that all webserver variables (ie. the values pulled
#     from the webserver cache) will be preceeded by webserver_ and those 
#     values associated with the engine(s) will be preceeded by engine_.
#------------------------------------------------------------------------------
sub configure
{
  my $interactive = (not INSTALL::conditional::boolean_value($pParameters->{non_interactive}));
  my $custom_install = INSTALL::conditional::boolean_value($pParameters->{custom_install});

  sub enabled {
    ($#_ == 0) || die "Invalid number of arguments.";
    
    my ($option) = @_;

    return($option ? "Enabled." : "Disabled.");
  }

  sub yes {
    ($#_ == 0) || die "Invalid number of arguments.";
    
    my ($option) = @_;

    return($option ? "Yes." : "No.");
  }

  sub exit_parent {
    ($#_ == 0) || die "Invalid number of arguments.";

    my ($pid) = @_;

    if ($interactive) {
      INSTALL::common::signal_spinner($pid);
    }
    else {
      wait;
    }    
    return(undef);
  }

  my %server_stats = @_;

  # Multi-threaded?
  my @models = ("multi-threaded", "multi-process");

  my $task_count = 5;
  my @default_ports = (1 + 1,($task_count * 10) + 1);
  my $model = 0;

  my %subst;
  
  if ((not $disable_multiprocess) && $interactive) {
    $model = INSTALL::query::user_menu
      ("ENGINE PARALLELISM MODEL",
       "Chili!Soft ASP may be run ineither multi-threaded or multi-process mode.  ".
       "Each of these have intrinisic qualities which make them more appropriate, ".
       "depending upon the situation.  Refer to your Chili!Soft ASP documentation ".
       "for descriptions of the virtues of each of the models.",
       "Run the engines in a multi-threadead mode",
       "Run the engines in a multi-process mode",
       "default=1");
    
    $model--;
  }

  if ($interactive) {
    INSTALL::query::user_menu
      ("CHILI!SOFT ASP ENGINE INSTALLATION",
       "You are currently configuring a new Chili!Soft ASP server which will be ".
       "associated with a webserver with the following information:\n \n".
       "  Webserver configuration file: $server_stats{conf}\n".
       "         Webserver port number: $server_stats{port}\n"
      );
  }

  ########## Install samples?
  my $install_samples = 1;

  if (exists $pParameters->{install_samples}) {
    $install_samples = INSTALL::conditional::boolean_value($pParameters->{install_samples});
  }

  if ($interactive && $custom_install) {
    $install_samples = query_boolean_default
      ("Would you like to enable samples on this server",$install_samples);
  }

  ########## Install docs?
  $install_docs = 1;

  if (exists $pParameters->{install_docs}) {
    $install_docs = INSTALL::conditional::boolean_value($pParameters->{install_docs});
  }

  if ($interactive && $custom_install) {
    $install_docs = query_boolean_default
      ("Would you like to enable documentation on this server",$install_docs);
  }

  ########## Automatic start on system boot (if available)?
  my $start_asp_onboot = 1;
  my $support_reboot = INSTALL::reboot::supported;

  if (exists $pParameters->{start_asp_onboot}) {
    $start_asp_onboot = INSTALL::conditional::boolean_value($pParameters->{start_asp_onboot});
  }

  if ($interactive && $custom_install && $support_reboot) {    
    $start_asp_onboot = query_boolean_default
      ("Would you like this Chili!Soft ASP server to start on system boot",
       $start_asp_onboot);
  }

  ########## Start ASP now?
  my $start_asp = 1;

  if (exists $pParameters->{start_asp}) {
    $start_asp = INSTALL::conditional::boolean_value($pParameters->{start_asp});
  }

  if ($interactive && $custom_install) {
    $start_asp = query_boolean_default
      ("Would you like to start this Chili!Soft ASP server",$start_asp);
  }

  ########## Do the actual configuration.
  if ($interactive) {
    print STDERR ("Configuring the Chili!Soft ASP server ... ");
  }

  my $pid = fork;

  if ($pid) {
    ########## Parent
    if (exists $pParameters->{asp_port}) {
      if (($pParameters->{asp_port} > 1024) && ($pParameters->{asp_port} < (32*1024)))
      {
	if (not 
	    INSTALL::services::validate_free_ports($pParameters->{asp_port},
						   $default_ports[$model]))
	{
	  note("The specified caspd / caspeng base port $pParameters->{asp_port} was",
	       "already in use.  However, due to the options specified, this port",
	       "has been allocated (perhaps inappropriately) for usage.  This may",
	       "cause the caspd / caspeng located in this port range to fail to",
	       "start.  Refer to the /etc/services file to find the conflict.");
	}

	$subst{caspeng::asp_port} = INSTALL::services::allocate_ports
	  ($pParameters->{asp_port},$default_ports[$model], "Chili!Soft ASP");	
      }
      else {
	$subst{caspeng::asp_port} = INSTALL::services::allocate_free_ports
	  (3000,$default_ports[$model], "Chili!Soft ASP");

	note("The specified asp_port=$pParameters->{asp_port} option was invalid.",
	     "Valid user port numbers range from 1024 to 32767.  As a result, the",
	     "following base port number was used: $subst{caspeng::asp_port}.");      
      }
    }
    else {
      $subst{caspeng::asp_port} = INSTALL::services::allocate_free_ports
	(3000,$default_ports[$model], "Chili!Soft ASP");
    }

    if ($models[$model] eq "multi-threaded") {
      $subst{caspeng::threaded} = 1;
    }
    else {
      $subst{caspeng::threaded} = 0;
    }
    
    $subst{caspeng::task_count} = $task_count;
    
    $subst{caspeng::server_dir} = 
      "$pGlobals->{asphome}/asp-$server_stats{type}-$subst{caspeng::asp_port}";
    
    my @server_path_components = split(m@/@,$subst{caspeng::server_dir});
    $subst{caspeng::server_name} = $server_path_components[$#server_path_components];    
    
    unless(mkdir($subst{caspeng::server_dir},0755)) {
      error("Unable to create server directory: $subst{caspeng::server_dir}");
      
      # Run cleanup code.
      return(exit_parent($pid));
    }
    
    if (system("/bin/cp $pGlobals->{asphome}/lib/caspd $subst{caspeng::server_dir}") ||
	system("/bin/cp $pGlobals->{asphome}/lib/caspeng $subst{caspeng::server_dir}") ||
	system("/bin/cp $pGlobals->{asphome}/lib/caspctrlb $subst{caspeng::server_dir}") ||
	system("/bin/cp $pGlobals->{asphome}/global_odbc.ini $subst{caspeng::server_dir}/odbc.ini"))
    {
      INSTALL::services::deallocate_ports($subst{caspeng::asp_port},$default_ports[$model]);
      system("rm -rf $subst{caspeng::server_dir}");
      error("Unable to copy the required server binaries into:",
		   "   $subst{caspeng::server_dir}");

      # Run cleanup code.
      return(exit_parent($pid));
    }
    
    ########## Template substitutions.
    if (not $install_samples) {
      $subst{caspeng::disable_samples} = "#";  # Comments out the samples lines.
    }
    else {
      $subst{caspeng::disable_samples} = ""; # Leave the samples lines.
    }
    
    if (not $install_docs) {
      $subst{caspeng::disable_docs} = "#";  # Comments out the docs line.
    }
    else {
      $subst{caspeng::disable_docs} = ""; # Leave the docs line.
    }
    
    INSTALL::substitute::template("stopcaspd",$subst{caspeng::server_dir},%subst);
    INSTALL::substitute::template("startcaspd",$subst{caspeng::server_dir},%subst);
    INSTALL::substitute::template("caspctrl",$subst{caspeng::server_dir},%subst);
    INSTALL::substitute::template("odbc.sh",$subst{caspeng::server_dir},%subst);
    INSTALL::substitute::template("casp.cnfg",$subst{caspeng::server_dir},%subst);
    INSTALL::substitute::template("setsqlnk",$subst{caspeng::server_dir},%subst);
    
    my @binary_files = 
      ("stopcaspd",
       "startcaspd",
       "caspctrl"
      );
    
    chmod(0755,map { "$subst{caspeng::server_dir}/$_" } @binary_files);
    
    ########## Install reboot scripts, if supported / enabled.
    if ($start_asp_onboot && $support_reboot) {
      INSTALL::reboot::install_script($subst{caspeng::server_name},
				      "$subst{caspeng::server_dir}/startcaspd",
				      "$subst{caspeng::server_dir}/stopcaspd",
				      "$subst{caspeng::server_dir}/caspctrl -status",
				      $CONST_StartPriority,$CONST_StopPriority);    
    }
    
    if ($start_asp) {
      if (system(" $subst{caspeng::server_dir}/startcaspd $debug_output")) {
	note("The Chili!Soft ASP engine failed to start.");
      }
    }
    
    ########## Update the installed_db.
    my $pIni = INSTALL::ini::open($pGlobals->{installed_db});
    if (!$pIni)
    {
      error("Could not open $pGlobals->{installed_db}.");

      # Run cleanup code.
      return(exit_parent($pid));
    }
    
    my $pSection = INSTALL::ini::insertSection($pIni, $subst{caspeng::server_dir});
    if (!$pSection)
    {
      error("Could not add section $subst{caspeng::server_dir} to $pGlobals->{installed_db}.");

      # Run cleanup code.
      return(exit_parent($pid));
    }
    
    $pSection->{engine_name} = $subst{caspeng::server_name};
    $pSection->{engine_dir} = "$pGlobals->{asphome}/$subst{caspeng::server_name}";
    $pSection->{engine_port} = $subst{caspeng::asp_port};
    $pSection->{engine_portcount} = $default_ports[$model];
    
    foreach(keys %server_stats) {
      $pSection->{"webserver_$_"} = $server_stats{$_};
    }
    
    $pGlobals->{status_text} .= 
      ("\nServer installed ($subst{caspeng::server_name}):\n".
       "  Associated webserver conf file: $server_stats{conf}\n".
       "  Associated webserver port: $server_stats{port}\n".
       "  Location: $subst{caspeng::server_dir}\n".
       "  Port: $subst{caspeng::asp_port}\n".
       "  Samples: ".enabled($install_samples)."\n".
       "  Documentation: ".enabled($install_docs)."\n".
       "  Automatic ASP start on system boot: ".enabled($start_asp_onboot && $support_reboot)."\n".
       "  ASP started: ".yes($start_asp)."\n".
       "  ASP start script: $subst{caspeng::server_dir}/startcaspd\n".
       "  ASP stop script: $subst{caspeng::server_dir}/stopcaspd\n".
       "  ASP general control script: $subst{caspeng::server_dir}/caspctrl\n"
      );
    
    if ($install_samples) {
      my $hostname = `hostname`;
      
      if (not $hostname) {
	$hostname = "<<your machine>>";
      }
      else {
	chomp $hostname;
      }
      
      $pGlobals->{status_text} .=
	"  Samples URL: http://$hostname:$server_stats{port}/caspsamp\n";
    }

    exit_parent($pid);
    return(INSTALL::ini::save($pIni, $pGlobals->{installed_db}));
  }
  else {
    ########## Child
    if ($interactive) {
      INSTALL::common::spinner_wait_signal;
    }
    exit(0);
  }
}

#------------------------------------------------------------------------------
# 
#==============================================================================
# Implementors note:
# (1) This method is guaranteed to receive not only the listing of all of the
#     cached webserver information, but also a single element of the hash
#     named 'engine_dir'.  This refers to the directories (ie. server_name),
#     of the Chili!Soft ASP engine being uninstalled.  It should be noted that
#     truncated version of engine_dir, which contains only its name, is also
#     passed via 'engine_name'.
# (2) The naming convention in %server_stats is different from that used during
#     the install.  As dictated in caspeng::configure, webserver_ is prepended
#     to all contained webserver information.
#------------------------------------------------------------------------------
sub uninstall
{
  my %server_stats = @_;
  my $interactive = (not INSTALL::conditional::boolean_value($pParameters->{non_interactive}));

  if (not exists $server_stats{engine_dir}) {
    return(error("Passed server statistics does not contain 'engine_dir' - uninstall."));
  }

  ########## Stop running engines.
  print STDERR ("Stopping engines of $server_stats{engine_name} ... ") if $interactive;

  my $pid = fork;
 
  if (not $pid) {
    system("$server_stats{engine_dir}/stopcaspd $debug_output");
    exit(0);
  }
  else {
    if ($interactive) {
      INSTALL::common::spinner_wait($pid);
      print "\n" if $interactive;
    }
    else {
      wait;
    }

    if ($?) {
      $result = 1;
    }
  }
  
  ########## Handle the relocation of the logfile.
  my $logfile;
  if (open(CASP_CNFG,"$server_stats{engine_dir}/casp.cnfg")) {
    foreach(<CASP_CNFG>) {
      if (/^\s*logfile\s*=\s*(.*)\s*$/) {
	$logfile = $1;
      }
    }
    close(CASP_CNFG);
  }

  ########## Ask for logfile relocation
#  if ((-f $logfile) && $interactive) {
#    my $answer = INSTALL::query::user
#      ("Would you like to keep a copy of your servers log file","terminal_mark=?",
#       "y","default=n");
#    
#    if ($answer eq "y") {
#      sub file_save_ok {
#	($filename) = @_;
#	
#	if ($filename ne "none") {
#	  my $parent = $filename;
#	  $parent =~ s@^(.*)/.*$@$1@;
#	  
#	  if ($parent eq "") {
#	    $parent = "/";
#	  }
#	  
#	  # Allow save only if valid parent directory, and the file does not
#	  # already exist.
#	  return((-d $parent) && (-w $parent) && (not -f $filename));
#	}
#	else {
#	  return(1);
#	}
#      }
#      
#      $answer = INSTALL::query::user_function
#	("Input the full path of the file you want to save the log to\n",
#	 *file_save_ok,
#	 "default=none",
#	 "invalid_choice=The specified path was not writable or that file already exists.");
#      
#      if ($answer ne "none") {
#	system("cp -f $logfile $answer");
#      }      
#    }
#  }
  
  if (-f $logfile) {
    chmod(0666,$logfile);
    unlink($logfile);
  }

  ########## Deallocate the used ports.
  INSTALL::services::deallocate_ports
    ($server_stats{engine_port},$server_stats{engine_portcount});

  ########## Remove all files in the server directory.
  system("rm -rf $server_stats{engine_dir}");

  ########## Uninstall the reboot script, if there is one.
  INSTALL::reboot::uninstall_script($server_stats{engine_name},
				    $CONST_StartPriority,$CONST_StopPriority);
  
  ########## Uninstall the installed_db section associated with this server.
  my $pIni = INSTALL::ini::open($pGlobals->{installed_db});
  if (!$pIni)
  {
    return(error("Could not open $pGlobals->{installed_db}."));
  }

  return(INSTALL::ini::deleteSection($pIni,$server_stats{engine_dir}) &&
	 INSTALL::ini::save($pIni,$pGlobals->{installed_db}));
}

1;
