########################################################################################
## CHILIASP INSTALLATION SCRIPT (for linux distributions)
##
## This does the main part of the install of Chili!Soft ASP.
########################################################################################

use strict;

use INSTALL::common;

use INSTALL::global;
use INSTALL::parameter;
use INSTALL::os;

my $CONST_Package = "main";
my $debug = 0;

STDOUT->autoflush(1);

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Internal functions
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#------------------------------------------------------------------------------
# All of the error reporting mechanisms take of a list of strings, and format
# them such that they fall on different lines.
#------------------------------------------------------------------------------
sub error {
  my $result = INSTALL::common::generate_error(@_);
  print STDERR "Package location: $CONST_Package\n";
  return($result);
}

sub trace
{
  if ($debug)
  {
    my $result = INSTALL::common::generate_report("trace: ",@_);
    print STDERR "Package location: $CONST_Package\n";
    return($result);
  }
}

sub note
{
  my $result = INSTALL::common::generate_note(@_);
  return($result);
}

sub report 
{
  my $result = INSTALL::common::generate_report("",@_);
  return($result);
}

my $pParameters = \%INSTALL::parameter::variables;

if ($> != 0) {
  error("caspi2.pl must execute as root.");
}

if (not defined $INSTALL::parameter::variables{asphome}) {
  error("asphome must be defined before this script will execute.");
  exit(1);
}

#==============================================================================
#------------------------ Default global values
#==============================================================================

INSTALL::common::set_standard_globals;

my $pGlobals = \%INSTALL::global::variables;
my $pOS = \%INSTALL::os::Details;

$pGlobals->{status_text} .=  
  "\n".
  "                     Chili!Soft ASP install summary\n".
  "======================================================================\n\n".
  "  installed version: $pGlobals->{casp_version}\n".
  "                 os: $pOS->{os}\n".
  "            release: $pOS->{release}\n".
  "               arch: $pOS->{hardware}\n".
  "        install dir: $pGlobals->{asphome}\n";

if ($pOS->{os} =~ /linux/i) {
  $pGlobals->{status_text} .=  
    ("       distribution: $pOS->{distribution}\n".
     "distribution flavor: $pOS->{distribution_flavor}\n");
}

#==============================================================================
#------------------------ Confirm pre-install requirements.
#==============================================================================

## Modify substitution permissions so they will take effect upon install.
my @binary_template_files = 
  ("chregsvr",
   "chregclass",
   "chregedit",
   "add_server",
   "delete_server",
   "uninstall",
   "list-webservers",
   "list-databases",
   "install_component",
   "apachectl",
   "vhostctl",

   # For internaldb, if in use.
   "start-internaldb","stop-internaldb"
  );

my @binary_files =
  ("$pGlobals->{install_dir}/unsetenv");

chmod(0755,
      (map { "$pGlobals->{template_dir}/$_" } @binary_template_files),
      @binary_files);

use INSTALL::ini;

my $interactive = (not INSTALL::conditional::boolean_value($pParameters->{non_interactive}));

# Create the log dir for the install.
if (not -d $pGlobals->{log_dir}) {
  mkdir($pGlobals->{log_dir},0755);
}

# Create the chili_ini repository directory.
if (not -d $pGlobals->{chili_ini_dir}) {
  my @subdirs = split(m@/@,$pGlobals->{chili_ini_dir});
  my $const = "";

  foreach(@subdirs) {
    $const .= "$_/";
      
    if (not -d $const) {
      mkdir($const,0755);
    }
  }
}

my $pIni=INSTALL::ini::open($pGlobals->{chili_ini});
my $override_install = INSTALL::conditional::boolean_value($pParameters->{override_install});

if (grep { $_ eq $pGlobals->{asphome} } INSTALL::ini::getSectionNames($pIni))
{

  if (not $override_install) {
    error("$pGlobals->{asphome} is already listed in the Chili!Soft ASP installation",
	  "database.  Make certain that the previous installation is removed",
	  "before reinstalling.");
    
    report("Advanced Users: ",
	   "If somehow this state was destroyed beyond repair, this fatal",
	   "error may be overcome by executing the install command:",
	   "  $pGlobals->{asphome}/install override_install=true");
    
    exit(1);
  }
  else {
    # Consider the data in chili.ini to be completely wrong.  Blow it away.
    INSTALL::ini::deleteSection($pIni,$pGlobals->{asphome});
  }
}

my $pSection = INSTALL::ini::insertSection($pIni,$pGlobals->{asphome});

if ($pSection) {
  # Save version information in the global section.

  $pSection->{casp_version} = $pGlobals->{casp_version};
  $pSection->{odbc_version} = $pGlobals->{odbc_version};
  INSTALL::ini::save($pIni,$pGlobals->{chili_ini});  
}

umask 0;
$ENV{CSREGISTRY} = "$pGlobals->{asphome}/chilicom/lib";

#==============================================================================
#------------------------ Confirm perl binary location.
#==============================================================================

use INSTALL::os;
use INSTALL::query;

# perl needs to be searched for and saved in a preceeding script.
sub confirm_perl_location 
{
  ($#_ == 0) || die "Invalid number of arguments to confirm_perl_location.";

  return(-x $_[0]);
}

$pGlobals->{perl} = $pOS->{perl};

if (not confirm_perl_location($pGlobals->{perl})) {
  $INSTALL::global::variables{perl} = INSTALL::query::user_function
    ("Setup was unable to locate your perl binary, enter it below\n",
     *confirm_perl_location);
}


#==============================================================================
#------------------------ Add global library path modifications if applicable.
#==============================================================================

if ($pOS->{os} =~ /linux/i) {
  if (not fork) {
    # Update the ld.so.conf to make chiliasp binaries immediately available.
    my $dupname = "/etc/ld.so.conf.pre-chiliasp";

    unlink($dupname);
    system("/bin/cp /etc/ld.so.conf $dupname");
    if (open(LD_SO_SRC,"<$dupname") &&
	open(LD_SO_CONF,">/etc/ld.so.conf")) 
    {
      unlink($dupname);
      while(<LD_SO_SRC>) {
	print LD_SO_CONF if ((not m@^$pGlobals->{asphome}$@) && (not m@^$pGlobals->{asphome}/lib$@));
      }
      
      print LD_SO_CONF "$pGlobals->{asphome}\n";
      print LD_SO_CONF "$pGlobals->{asphome}/lib\n";
      close(LD_SO_CONF);
      
      my $ldconfig = "/sbin/ldconfig";
      
      if (-x "/bin/ldconfig") {
	$ldconfig = "/bin/ldconfig";
      }
      
      system("$ldconfig");
    }

    close(LD_SO_CONF);
    close(LD_SO_SRC);

    exit(0);
  }
  $pGlobals->{status_text} .=
    "\n/etc/ld.so.conf was updated with Chili!Soft ASP library paths.\n";
}

#==============================================================================
#------------------------ Generate pre-install / uninstall script.
#==============================================================================

use INSTALL::substitute;

my $install_script = "$pGlobals->{asphome}/INSTALL/install";

if (not INSTALL::substitute::template("uninstall","$pGlobals->{asphome}"))
{
  error("Unable to generate uninstall script in $pGlobals->{asphome}");
  exit(1);
}

open (INSTALL,">$install_script") || 
  die "Unable to create $install_script script.";

print INSTALL ("#!/bin/sh\n\n",
	       "$pGlobals->{perl} -I$pGlobals->{install_dir} ",
	       "$pGlobals->{asphome}/INSTALL/caspi2.pl \"asphome=$pGlobals->{asphome}\" \"\$\@\"\n");
close(INSTALL);
chmod(0755,$install_script);

if (not INSTALL::substitute::template("install_component","$pGlobals->{asphome}/INSTALL")) {
  note("Failed to generate the install_component script in $pGlobals->{asphome}.");
}

#==============================================================================
#------------------------ Check for LICENSE.LIC file.
#==============================================================================

use INSTALL::conditional;

if ($interactive && (not -f "$pGlobals->{asphome}/LICENSE.LIC")) {
  sub file_exists
  {
    my ($filename) = @_;

    return(($filename eq "none") || (INSTALL::conditional::file_exists($filename)));
  }

 REQUERY_LICENSE:
  INSTALL::query::user_menu
    ("LICENSE FILE INSTALLATION",
     "Chili!Soft ASP will not function without a valid license file in the ".
     "$pGlobals->{asphome} directory (prior to execution).  If you do not wish to ".
     "install a license file now, you may proceed with the rest of the ".
     "installation.  It should be stressed, however, that without a license ".
     "file Chili!Soft ASP will be non-functional. This file may be obtained in ".
     "the following ways:\n".
     "1.  If this product was downloaded via the web, you should have received\n".
     "    email instructions on downloading and saving the license file,\n".
     "    locally; in this case enter the full path of the local copy of the\n".
     "    license file.\n".
     "2.  If you are upgrading from any version other than 3.5.01, and you\n".
     "    do not have a backup copy of your LICENSE.LIC file, you will\n".
     "    need to contact Chili!Soft support to obtain a new license file.\n".
     "    Refer to the README file for information on contacting support.");

  
  my $answer = INSTALL::query::user
    ("Would you like to install a provided license file",
     "terminal_mark=?", "default=y", "n");

  if ($answer eq "y") {
    report("");
    note("If you wish to exit out of the license file installer, type none.");
    report("");
    $answer = INSTALL::query::user_function
      ("Enter the full path and filename of your license file\n",
       *file_exists,
       "default=none",
      "invalid_choice=That file does not exist or is not a valid license file.");

    if ($answer ne "none") {
      system("/bin/cp $answer $pGlobals->{asphome}/LICENSE.LIC");
      chmod(0644,"$pGlobals->{asphome}/LICENSE.LIC");
      
      $pGlobals->{status_text} .=
	"\nLicense file $answer was installed as $pGlobals->{asphome}/LICENSE.LIC.\n";
    }
    else {
      goto REQUERY_LICENSE;
    }
  }
  else {
    $pGlobals->{status_text} .=
      "\nNo license file was installed.\n";
  }
}

#==============================================================================
#------------------------ Install chiliasp user.
#==============================================================================

use INSTALL::user;

my $chiliasp_installed = 1;

########## We used to ask for confirmation of keeping the current settings
########## of the chiliasp user if they were changed from 'stock' settings.
########## For security reasons, we now blow away any current chiliasp user
########## from the box and install a new user.

if (INSTALL::user::exists("chiliasp")) {
  INSTALL::user::del("chiliasp");
}

unless (INSTALL::user::add("chiliasp","psailihc","/dev/null",
			   undef,undef,"Chili!ASP Database User")) 
{
  $chiliasp_installed = 0;
}

if (not $chiliasp_installed) {
  error("Unable to create chiliasp user. We cannot locate the tools",
	"necessary to perform this installation.  This might be due to",
	"several things, but most likely is due to either:",
	"   (1) You are on a distribution our product does not support.",
	"   (2) You are on a heavily modified distribution we do support.",
	"   (3) Some configuration setting(s) on your machine are",
	"       disabling some set of needed scripts.");
  note("This user must be in place before the installation may proceed.  Use",
       "whatever mechanism that exists on your machine to create a user such",
       "that the following information is true:",
       "  Username: chiliasp",
       "  Password: psailihc",
       "  Login Shell: /dev/null",
       "",
       "After this has been made true, execute $install_script."
      );
  exit(1);
}
else { 
  $pGlobals->{status_text} .=
    "\nA system-wide chiliasp user was added / modified.\n";
}

#==============================================================================
#------------------------ Wait for ld.so.conf child to resume.
#==============================================================================

wait;

#==============================================================================
#------------------------ Configure internaldb.
#==============================================================================


use INSTALL::internaldb;

if (not INSTALL::internaldb::configure) {
  error("Unable to complete internal database installation.  Internal",
	"database samples will be non-functional.");
}


#==============================================================================
#------------------------ Configure global items (ie. registry / root scripts).
#==============================================================================

## Substitutions
my @functional_scripts =
  ("list-webservers",
   "list-databases",
   "settings.reg",
   "apachectl",
   "vhostctl"
  );

foreach(@functional_scripts) {
  INSTALL::substitute::template($_,$pGlobals->{install_dir});
}

my @subst_files =
  ("chsetup.sh",
   "chregsvr",
   "chregclass",
   "chregedit",
   "add_server",
   "delete_server",
   "javasetup.sh",
  );

foreach(@subst_files) {
  INSTALL::substitute::template($_,$pGlobals->{asphome});
}

my $regedit = "$pGlobals->{asphome}/chregedit";
my $regsvr = "$pGlobals->{asphome}/chregsvr";

if (not fork) {
  $pGlobals->{register_summary} .= "\nEntering default registry settings: Success.\n";
  system("$regedit $pGlobals->{settings_reg} > $pGlobals->{summary_log} 2>&1");
  
  my @register_libs =
    ((map { "$pGlobals->{asphome}/chilicom/lib/$pOS->{ostag}_optimized/lib$_.so" }
      ("atl",
       "vbscript",
       "jscript")),
     
     (map { "$pGlobals->{asphome}/lib/lib$_.so" }
      (
       # Core objects.
       "chiliado",
       "chiliscrrun",
       "calc",
       "chilicap",
       "chiliasp",

       # IcePack libraries (non-SpicePack).
       "chiliadrotator",
       "chilicontentrotator",
       "chilinextlink",
       "chilitools",
       "counter",
       "myinfo"
      )),
     
     (map { "$pGlobals->{asphome}/bean/cbnative/lib$_.so" }
      ("cbnative"))
    );
  
  foreach(@register_libs) {
    my $libname = $_;
    
    if (m@([^/]*)$@) {
      $libname = $1;
    }
    
    $pGlobals->{register_summary} .= "\nRegistering $libname: ";
    
    if (not system("$regsvr $_ >> $pGlobals->{install_log} 2>&1")) {
      $pGlobals->{register_summary} .= "Success.";
    }
    else {
      $pGlobals->{register_summary} .= "Failed.";
    }
    
    $pGlobals->{register_log} .= "\n";
  }
  
  if (open(REGISTER,">$pGlobals->{register_log}")) {
    print REGISTER "$pGlobals->{register_summary}\n";
    close(REGISTER);
  }
  
  exit(0);
}

#==============================================================================
#------------------------ Configure initial webserver.
#==============================================================================
use INSTALL::webserver;

if ((not exists $pParameters->{install_webserver}) ||
    INSTALL::conditional::boolean_value($pParameters->{install_webserver}))
{
  # Add a new webserver if either they did not specify an install_webserver
  # option, or install_webserver was set to a true value.  Otherwise, 
  # do not install a webserver.

  INSTALL::webserver::build_server_cache;
  INSTALL::webserver::configure;
}

#==============================================================================
#------------------------ Configure any available components.
#==============================================================================

if (opendir(DIR,$pGlobals->{components_dir})) {
  my @components = readdir(DIR);
  closedir(DIR);

  if ($#components >= 0) {
    $pGlobals->{status_text} .=
      "\nComponent installation:\n";
  }

  foreach(@components) {
    next if (($_ eq ".") || ($_ eq ".."));

    my $component_dir = "$pGlobals->{components_dir}/$_";

    if (-f "$component_dir/install.pl") {
      if (system("$pGlobals->{perl} -I$pGlobals->{asphome} -I$component_dir $component_dir/install.pl \"component_dir=$component_dir\" $pGlobals->{execute_params}")) 
      {
	$pGlobals->{status_text} .=
	  "  $_: Failure or Not Installed.\n";
      }
      else {
	$pGlobals->{status_text} .=
	  "  $_: Success.\n";
      }
    }
    elsif (-d $component_dir) {
      note("Discovered component '$_' without an installation script.");
    }
  }
}

#==============================================================================
#------------------------ Write out summary log.
#==============================================================================

if (open(SUMMARY,">$pGlobals->{summary_log}")) {
  print SUMMARY "$pGlobals->{status_text}\n";
  close(SUMMARY);
}

#==============================================================================
#------------------------ Installation complete.
#==============================================================================

if ($interactive) {
  INSTALL::query::user_menu
    ("INSTALLATION COMPLETE",
     "The Chili!Soft ASP installation has completed successfully.  Refer to the ".
     "summary file listed below for more details on using Chili!Soft ASP.  Thanks ".
     "again for choosing Chili!Soft."
    );

  print "Summary file: $pGlobals->{summary_log}\n";
}

exit(0);
