package INSTALL::component;
use INSTALL::ini;

$system_database = "/var/opt/casp/chili.ini";

$enable_debug = 0;
$debug_output = $enable_debug ? "" : "> /dev/null 2>&1";

#==============================================================================
# Miscelaneous / Helpers
#==============================================================================

sub error 
{
  print STDERR "Error: ", @_, "\n";
  return(0);
}


sub encode_text {
  ($#_ == 2) || die "Invalid number of arguments to INSTALL::component::encode_text";
  my ($text, $orig_char, $new_char) = @_;

  my $orig_char_pattern = quotemeta($orig_char);
  my $new_char_pattern = quotemeta($new_char);

  $text =~ s/$new_char_pattern/$new_char$new_char/g;
  $text =~ s/$orig_char_pattern/$new_char/g;

  return($text);
}

sub unencode_text {
  ($#_ == 2) || die "Invalid number of arguments to INSTALL::component::unencode_text";
  ($text, $orig_char, $new_char) = @_;

  my $orig_char_pattern = quotemeta($orig_char);
  my $new_char_pattern = quotemeta($new_char);
 
  $text =~ s/^$new_char_pattern([^$new_char_pattern])/$orig_char$1/g;
  $text =~ s/([^$new_char_pattern])$new_char_pattern([^$new_char_pattern])/$1$orig_char$2/g;
  $text =~ s/$new_char_pattern$new_char_pattern/$new_char/g;

  return($text);
}

sub pHash_to_text {
  ($#_ == 0) || die "Invalid number of arguments to INSTALL::component::pHash_to_text";

  my $pHash = $_[0];
  my @sections = 
    map
    {
      my $key = $_;
      encode_text(encode_text($key,"=","~"),",",";") . "=" .
	encode_text($pHash->{$key},",",";");
    } keys %{$pHash};

  return(join(",",@sections));
}

sub text_to_pHash {
  ($#_ == 0) || die "Invalid number of arguments to INSTALL::component::text_to_pHash";
  
  my ($section_text) = @_;  
  my @sections = split(/\s*,\s*/,$section_text);
  my %hash;

  foreach (@sections) {
    /([^=]*)=(.*)/;
    $hash{unencode_text(unencode_text($1,",",";"),"=","~")} = encode_text($2,",",";");
  }

  return(\%hash);
}

sub encode_name {
  ($#_ == 0) || die "Invalid number of arguments to INSTALL::component::encode_name";
  
  my ($name) = @_;  
  return(encode_text($name," ","_"));
}

sub unencode_name {
  ($#_ == 0) || die "Invalid number of arguments to INSTALL::component::unencode_name";
  
  my ($name) = @_;  
  return(unencode_text($name," ","_"));
}

#------------------------------------------------------------------------------
#==============================================================================
# Exposed interfaces
#==============================================================================
#------------------------------------------------------------------------------

# Returns non-zero on success.
sub register 
{
  #---------------------------------------------------------------------------
  # Parameter:                  Default value:                  Optional:
  #---------------------------------------------------------------------------
  # name                        <none>                          N
  # installdir                  <none>                          N
  # hash                        <empty>                         Y
  
  my $length = $#_ + 1;
  if ($length < 2) {
    return(error("Invalid number of parameters to INSTALL::component::register [",
                 $length, "]."));
  }    

  my ($name, $installdir, %hash) = @_;
  
  if (!$name) {
    return error("INSTALL::component::register - name parameter cannot be empty.");
  }
  
  my $pIni = INSTALL::ini::open($system_database);
  if (!$pIni) {
    return error("INSTALL::component::register - Could not update '$system_database'.");
  }

  my $pSection = INSTALL::ini::openSection($pIni, "$installdir");
  if (!$pSection)
  {
    return error("INSTALL::component::register - Could not find section '$installdir'.");
  }

  my $key = "component_" . encode_name($name);

  if (exists $pSection->{$key}) {
    my $pCurHash = text_to_pHash($pSection->{$key});
    foreach (keys %{$pCurHash}) {
      if (not exists $hash{$_}) {
	$hash{$_} = $pCurHash->{$_};
      }
    }
  }

  $pSection->{$key} = pHash_to_text(\%hash);
  INSTALL::ini::save($pIni, $system_database);
  
  return(1);
}

# Returns non-zero on success.
sub unregister 
{
  #---------------------------------------------------------------------------
  # Parameter:                    Default value:                  Optional:
  #---------------------------------------------------------------------------
  # name                          <none>                          N
  # installdir                    <none>                          N
    
  my $length = $#_ + 1;
  if ($length != 2) {
    return(error("Invalid number of parameters to INSTALL::component::unregister [",
		 $length, "]."));
  }    

  my ($name, $installdir) = @_;  

  # find name in section
  my $pIni = INSTALL::ini::open($system_database);
  if (!$pIni)
  {
    return error("INSTALL::component::unregister - Could not update '$system_database'.");
  }
  my $pSection = INSTALL::ini::openSection($pIni, "$installdir");
  if (!$pSection)
  {
    return error("INSTALL::component::unregister - Could not find section '$installdir'.");
  }

  my $key = "component_" . encode_name($name);
  my $result = 0;

  if (exists $pSection->{$key}) {
    delete $pSection->{$key};
    $result = 1;
  }

  INSTALL::ini::save($pIni, $system_database);
  
  return($result);
}

# Returns non-zero on success.
sub list
{
  #---------------------------------------------------------------------------
  # Parameter:                    Default value:                  Optional:
  #---------------------------------------------------------------------------
  # installdir                    <none>                          N
    
  my $length = $#_ + 1;
  if ($length != 1) {
    return(error("Invalid number of parameters to INSTALL::component::lists [",
		 $length, "]."));
  }
  
  my ($installdir) = @_;
  
  my $pIni = INSTALL::ini::open($system_database);
  if (!$pIni)
  {
    error("INSTALL::component::lists - Could not update '$system_database'.");
    return @error;
  }
  my $pSection = INSTALL::ini::openSection($pIni, "$installdir");
  if (!$pSection)
  {
    error("INSTALL::component::lists - Could not find section '$installdir'.");
    return @error;
  }

  my @component_keys = grep { /component_/ } keys %{$pSection};
  my @components = map { s/^component_//; unencode_name($_) } @component_keys;

  return(@components);
}       

# Returns a reference to a hash which contains all of the values associated 
# with the specified component (or undef, on failure).
sub open
{
  #---------------------------------------------------------------------------
  # Parameter:                    Default value:                  Optional:
  #---------------------------------------------------------------------------
  # name                          <none>                          N
  # installdir                    <none>                          N
    
  my $length = $#_ + 1;
  if ($length != 2) {
    return(error("Invalid number of parameters to INSTALL::component::open [",
		 $length, "]."));
  }
  
  my ($name, $installdir) = @_;
  
  my $pIni = INSTALL::ini::open($system_database);
  if (!$pIni)
  {
    error("INSTALL::component::open - Could not update '$system_database'.");
    return @error;
  }
  my $pSection = INSTALL::ini::openSection($pIni, "$installdir");
  if (!$pSection)
  {
    error("INSTALL::component::open - Could not find section '$installdir'.");
    return @error;
  }

  my $key = "component_" . encode_name($name);

  if (exists $pSection->{$key}) {
    return(text_to_pHash($pSection->{$key}));
  }
  else {
    return(undef);
  }
}

# Returns non-zero on success.  It should be noted that the third parameter 
# (ie. <pHash>) is a pointer to a hash.  Presumably this is the value returned
# from open.
sub save {
  #---------------------------------------------------------------------------
  # Parameter:                    Default value:                  Optional:
  #---------------------------------------------------------------------------
  # name                          <none>                          N
  # installdir                    <none>                          N
  # pHash                         <none>                          Y
  
  my $length = $#_ + 1;

  if ($length < 2) {
    return(error("Invalid number of parameters to INSTALL::component::save [",
		 $length, "]."));
  }

  my ($name,$installdir,$pHash) = @_;
  
  my $pIni = INSTALL::ini::open($system_database);
  if (!$pIni)
  {
    error("INSTALL::component::save - Could not update '$system_database'.");
    return @error;
  }

  my $pSection = INSTALL::ini::openSection($pIni, "$installdir");
  if (!$pSection)
  {
    error("INSTALL::component::save - Could not find section '$installdir'.");
    return @error;
  }

  my $key = "component_" . encode_name($name);

  $pSection->{$key} = pHash_to_text($pHash);
 
  INSTALL::ini::save($pIni, $system_database);

  return(1);
}

# Returns non-zero on success.  It should be noted that the third parameter 
# (ie. <hash>) is an actual hash. This method is similar to save
# with the exception that the current hash values are merged with the new 
# ones.  Current values will be overwritten by new values of the same key.
sub merge {
  #---------------------------------------------------------------------------
  # Parameter:                    Default value:                  Optional:
  #---------------------------------------------------------------------------
  # name                          <none>                          N
  # installdir                    <none>                          N
  # hash                          <none>                          Y
  
  my $length = $#_ + 1;

  if ($length <= 2) {
    return(error("Invalid number of parameters to INSTALL::component::merge [",
		 $length, "]."));
  }

  my ($name,$installdir,%hash) = @_;
  
  my $pIni = INSTALL::ini::open($system_database);
  if (!$pIni)
  {
    error("INSTALL::component::merge - Could not update '$system_database'.");
    return @error;
  }

  my $pSection = INSTALL::ini::openSection($pIni, "$installdir");
  if (!$pSection)
  {
    error("INSTALL::component::merge - Could not find section '$installdir'.");
    return @error;
  }

  my $key = "component_" . encode_name($name);

  if (exists $pSection->{$key}) {
    my $pCurHash = text_to_pHash($pSection->{$key});
    foreach (keys %{$pCurHash}) {
      if (not exists $hash{$_}) {
	$hash{$_} = $pCurHash->{$_};
      }
    }
  }

  $pSection->{$key} = pHash_to_text(\%hash);
 
  INSTALL::ini::save($pIni, $system_database);

  return(1);
}

1;
