package INSTALL::ini;

use INSTALL::common;

my $CONST_Package="INSTALL::ini";

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Internal functions
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#------------------------------------------------------------------------------
# All of the error reporting mechanisms take a list of strings, and format
# them such that they fall on different lines.
#------------------------------------------------------------------------------
sub error {
  my $result = INSTALL::common::generate_error(@_);
  print STDERR "Package location: $CONST_Package\n";
  return($result);
}

sub note {
  my $result = INSTALL::common::generate_note(@_);
  print STDERR "Package location: $CONST_Package\n";
  return($result);
}

sub trace
{
  if ($debug)
  {
    my $result = INSTALL::common::generate_report("trace: ",@_);
    print STDERR "Package location: $CONST_Package\n";
    return($result);
  }
}

sub report 
{
  my $result = INSTALL::common::generate_report("",@_);
  print STDERR "Package location: $CONST_Package\n";
  return($result);
}


#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# External functions
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#------------------------------------------------------------------------------
# Open an ini file, returning a reference to an internally maintained structure
# which represents the ini file.  The specified ini file need not exist.  If
# it does not exist, a new ini file will be created.
#------------------------------------------------------------------------------
sub open {
  ($#_ == 0) || die "Error: Invalid number of parameters (must be 1).";
  my ($fileName) = @_;

  if (stat($fileName)) {
    return(load($fileName));
  }
  elsif (open(FILE,">$fileName")) {
    close(FILE);
    return(load($fileName));
  }
  else {
    return(error("Unable to open $fileName for writing - open.\nReason: $!"));
  }
}

#------------------------------------------------------------------------------
# Opens an existing ini file.  If the file does not exist, an error will be 
# returned.  Otherwise, a reference to an ini structure will be returned.
# NOTE: The functionality of load is identical to open, except for the 
# requirement of the ini file to already exist. 
#------------------------------------------------------------------------------
sub load {
  ($#_ == 0) || die "Error: Invalid number of parameters (must be 1).";

  my ($fileName) = @_;

  unless (open(FILE,"<$fileName")) {
    return(error("Unable to open $fileName for reading - load\nReason: $!."));
  }
  else {
    my $line;
    my %result;

    my %newSection;
    my $newName;
    my @section_order;
    my %comments;

    while($line = <FILE>) {
      if ($line =~ /^\s.*$/) {
	# Ignore empty lines.
	next;
      }
      elsif ($line =~ /^\s*\#\s*(.*)$/) {
	if ($newName) {
	  push @{$comments{$newName}},$1;
	}
	else {
	  # Global comment name = [
	  push @{$comments{"["}},$1;
	}
	next;
      }
      elsif ($line =~ /^\s*\[([^\]]*)\]\s*$/) {
	if ($newName) {
	  my %insert=%newSection;
	  $result{$newName} = \%insert;
	  $newName = "";
	  %newSection = ();
	}

	$newName = $1;
	push @section_order, $newName;
      }
      elsif ($line =~ /^\s*([^=]+)\s*=\s*(.*)$/) {
	my $key = $1;
	my $value = $2;
	
	# For backware compatibility reasons, if there is a double-quote at 
	# the head and tail of a value, strip them.	
	$value =~ s@^\"(.*)\"$@$1@;

	$newSection{$key} = $value;
      }
      elsif ($line =~ /^\s*([^=]+)\s*$/) {
	$newSection{$1} = "";
      }
      else {
	note("Skipping garbage line: $line");
      }
    }

    if ($newName) {
      $result{$newName} = \%newSection;
    }

    $result{"]section_order"} = \@section_order;
    $result{"]comments"} = \%comments;

    close(FILE);
    return \%result;
  }
}

#------------------------------------------------------------------------------
# Saves the information contained in the specified passed ini reference into
# the specified file.  
#------------------------------------------------------------------------------
sub save {
  ($#_ = 2) || die "Error: Invalid number of parameters (must be 2).";

  my ($pIniData,$fileName) = @_;
  my %iniData = %{$pIniData};

  unless (open(FILE,">$fileName")) {
    return(error("Unable to open $fileName for writing - save.\nReason: $!"));
  }
  else {
    # Handle external comments.
    my %comments;
    
    if (exists $iniData{"]comments"}) {
      %comments = %{$iniData{"]comments"}};
    }

    if (exists $comments{"["}) {
      my @external_comments = @{$comments{"["}};
      
      foreach(@external_comments) {
	print FILE "# $_\n";
      }

      print FILE "\n";
    }
        
    my @section_order;

    if (exists $iniData{"]section_order"}) {
      @section_order = @{$iniData{"]section_order"}};
    }
    else {
      @section_order = getSectionNames($pIniData);
    }
    
    # For each section, print it out.
    foreach (@section_order) {
      my $section_name = $_;

      print FILE "[$section_name]\n";
      
      my %records = %{$iniData{$_}};
      my @recordKeys = keys %records;

      foreach (@recordKeys) {
	my $record = $records{$_};
	print FILE "$_=$record\n";
      }

      # Handle internal comments.
      if (exists $comments{$section_name}) {
	my @internal_comments = @{$comments{$section_name}};

	foreach(@internal_comments) {
	  print FILE "# $_\n";
	}
      }

      print FILE "\n";
    }
      
    close(FILE);
    return(1);
  }
}

#------------------------------------------------------------------------------
# Insert a new section into the passed ini structure.  If the named section 
# already exists, the return value will be that section.  It is thus impossible
# to store more than one section of the same name in an ini structure.  Upon
# success, the return value will be a reference to a section structure (a hash).
# It should be noted that changes made to this structure will result in 
# alterations to the passed ini structure (by design).  
#------------------------------------------------------------------------------
sub insertSection {
  ($#_ == 1) || die "Error: Invalid number of parameters (must be 2).";

  my ($pIniData,$section) = @_;

  if (exists $pIniData->{$section}) {
    return(openSection($pIniData,$section));
  }
  else {
    my %newSection;
    $pIniData->{$section} = \%newSection;

    if (not grep { $section eq $_ } @{$pIniData->{"]section_order"}}) {
      push @{$pIniData->{"]section_order"}}, $section;
    }

    return(\%newSection);
  }
}

#------------------------------------------------------------------------------
# Delete the specified section from the passed ini structure reference. The
# reutnr value is undef, if the operation failed, or non-undef on success.  It
# should be noted that the passed ini reference (or rather the ini structure to
# which that refers), will be altered via this method.
#------------------------------------------------------------------------------
sub deleteSection {
  ($#_ == 1) || die "Error: Invalid number of parameters (must be 2).";

  my ($pIniData,$section) = @_;

  if (exists $pIniData->{$section}) {
    my $pRemoved = $pIniData->{$section};
    delete $pIniData->{$section};
    
    if (grep { $section eq $_ } @{$pIniData->{"]section_order"}}) {
      my @newOrder;
      foreach(@{$pIniData->{"]section_order"}}) {
	if ($section ne $_) {
	  push @newOrder, $_;
	}
      }
      
      $pIniData->{"]section_order"} = \@newOrder;
    }

    return($pRemoved);
  }
  else {
    return(undef);
  }
}

#------------------------------------------------------------------------------
# Retrieve a list of names from the passed ini structure reference.
#------------------------------------------------------------------------------
sub getSectionNames {
  ($#_ == 0) || die "Error: Invalid number of parameters (must be 1).";

  my ($pIniData) = @_;
  my %iniData = %{$pIniData};

  my @result;
  foreach(keys %iniData) {
    # Strip special fields.
    if (not /\]/) {
      push @result,$_;
    }
  }

  return(@result);
}

#------------------------------------------------------------------------------
# Open the specified section if the section exists.  If the section does not 
# exist failure will be returned.  This method is identical to insertSection,
# except that if the section does not already exist, failure will be returned.
#------------------------------------------------------------------------------
sub openSection {
  ($#_ == 1) || die "Error: Invalid number of parameters (must be 2).";

  my ($pIniData,$section) = @_;

  if (exists $pIniData->{$section}) {
    return($pIniData->{$section});
  }
  else {
    return(undef);
  }
}

#------------------------------------------------------------------------------
# Each section also has associated with it a comments section.  This section
# of comments will be appended to the end of the section data as # <comment> 
# ... for each comment contain in the list.  When read from the ini file in 
# question, they will be output in the order read in.  This method returns a
# reference to an array.  In this way, by changing this array, you will be
# changing the ini data.  If the section named does not already exist, undef
# will be returned (see insertComments).
#------------------------------------------------------------------------------
sub openComments {
  ($# == 0) || ($# == 1) || die "Invalid number of arguments.";

  my ($pIniData,$section) = @_;

  if (not $section) {
    $section = "[";
  }
  elsif (not exists $pIniData->{$section}) {
    return(undef);
  }
  else {
    if (not exists $pIniData->{"]comments"}) {
      my %newComment;
      $pIniData{"]comments"} = \%newComment;
    }
    
    if (not exists $pIniData->{"]comments"}->{$section}) {
      my @newComments;
      $pIniData->{"]comments"}->{$section} = \@newComments;
    }

    return($pIniData->{"]comments"}->{$section});
  } 
}

#------------------------------------------------------------------------------
# Identical to openComments except that if the specified section does not exist
# it will be added.
#------------------------------------------------------------------------------
sub insertComments {
  ($# == 0) || ($# == 1) || die "Invalid number of arguments.";

  my ($pIniData,$section) = @_;

  if (not $section) {
    $section = "[";
  }
  elsif (not exists $pIniData->{$section}) {
    my %newSection;
    $pIniData->{$section} = \%newSection;
  }

  return(openComments($pIniData,$section));
}

1;
