package INSTALL::query;

use INSTALL::conditional;

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Internal functions
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#------------------------------------------------------------------------------
# Confirms / enforces that each item of the list is only listed once in the
# output.
#------------------------------------------------------------------------------
sub squeeze_list 
{
  my (@passed_list) = @_;
  my @list;

  foreach (@passed_list) {
    my $item = $_;
    if (not grep { /^$item$/ } @list) {
      push @list, $_;
    }
  }

  return(@list);
}

#------------------------------------------------------------------------------
# Constructs a string which is the result of repeating <repeat_text>, <count>
# times.
#------------------------------------------------------------------------------
sub repeat_text 
{
  ($#_ == 1) || die "Invalid number of arguments.";

  my ($repeat_text, $count) = @_;

  my $i;
  my $result = "";
  for($i = 0; $i < $count; $i++) {
    $result .= $repeat_text;
  }

  return($result);
}

#------------------------------------------------------------------------------
# Places <text> into a justified region of <justify_length>, surrounded by 
# <terminal_left> and <terminal_right> characters.  The <text> may span several
# lines.
#------------------------------------------------------------------------------
sub justify_text 
{
  sub justify_text_block {
    ($#_ == 3) || die "Invalid number of arguments.";
    
    my ($text, $justify_length, $terminal_left, $terminal_right) = @_;
    my $length = 0;
    my $result;
    
    $justify_length -= length($terminal_left) + length($terminal_right);
    
    while($text) {
      if ($text =~ /(\s*\S*\s*)/) {
	my $word = $1;
	my $pattern = quotemeta($word);
	
	$text =~ s/$pattern//;
	
	if ($length == 0) {
	  $result .= $terminal_left;
	}
	
	chomp $word;
	my $word_length = length($word);
	
	if ($word_length > $justify_length) {
	  if ($length) {
	    $result .= repeat_text(" ",$justify_length - $length) .
	      "$terminal_right\n$terminal_left$word$terminal_right\n";
	  }
	  else {
	    $result .= "$word$terminal_right\n";
	  }
	  $length = 0;
	}
	elsif (($word_length + $length) >= $justify_length) {
	  $result .=  repeat_text(" ",$justify_length - $length) .
	    "$terminal_right\n$terminal_left" . $word;
	  $length = $word_length;
	}
	else {
	  $result .= $word;
	  $length += $word_length;
	}
      }
    }
    
    if ($length && ($length < $justify_length)) {
      $result .= repeat_text(" ",$justify_length - $length)."$terminal_right\n";
    }
    
    return($result);
  } 

  ($#_ == 3) || die "Invalid number of arguments.";
  
  my ($text, $justify_length, $terminal_left, $terminal_right) = @_;

  my @newline_text = split(m@\n@,$text);
  my $result = "";

  foreach(@newline_text) {
    $result .= justify_text_block($_,$justify_length,$terminal_left,$terminal_right);
  }

  return($result);
}

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# External functions
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#------------------------------------------------------------------------------
# This method queries the user such that the resultant value will be passed to
# a user defined function <query_function>.  If this function returned success,
# then the value is accepted and the result is returned.  If the function 
# returns failed, $options{invalid_choice} will be displayed to the user and 
# the user will be requeried.  
# 
# Parameter 1 (<query_string>) is the string which will be prompted to the user
# to answer.
#
# Parameter 2 (<query_function>) is the user defined function which will be 
# queried for answer validity.
#
# Parameter 3 - n (<passed_valid_values>) are optional parameters which, among
# other things, provides a list of options which will be listed alongside the
# <query_string> as valid options.  It should be noted that these options must
# pass <query_function> lest a problem will occur.  These parameters may also
# be keyed option specifiers.  For example, "terminal_mark=?" will cause the
# terminal character at the end of <query_string> to be '?'.  The following
# list of options exists:
#  terminal_mark=<option> (default value: ':')
#    - changes the character(s) which succeed the <query_string>.  
#  
#  invalid_choice=<option> (default value: "Invalid choice")
#    - changes the string which gets displayed when the user inputs an invalid
#      option.  If <option> = "", no 'Invalid choice' message will be 
#      displayed.  Instead, it is assumed that the user function will handle
#      this.
#       
#  default=<option> (default value: none)
#    - changes the default behavior of hitting return without inputing any 
#      data.  If a user hits return, with "default=foo", the resulting return
#      will be interpreted as if the user had typed foo and then hit return.
#
#  case_sensitive=[yes|no] (default value: no)
#    - specifies whether or not matches are case sensitive.
#------------------------------------------------------------------------------
sub user_function
{
  my ($query_string, $query_function, @passed_valid_values) = @_;

  my %options = (
		 terminal_mark => ":",
		 invalid_choice => "Invalid choice",
		 break_key => "break-out"
		);
  my @valid_values;

  foreach (@passed_valid_values) {
    if (/default\s*=\s*(.*)/si) {
      (not exists $options{default_value}) || die "More than one default value in query_user.";
      $options{default_value} = $1;
      push @valid_values,$1;
    }
    elsif (/terminal_mark\s*=\s*(.*)/si) {
      $options{terminal_mark}=$1;
    }
    elsif (/invalid_choice\s*=\s*(.*)/si) {
      $options{invalid_choice}=$1;
    }
    elsif (/case_sensitive\s*=\s*(.*)/si) {
      $options{case_sensitive} = $1;
    }
    else {
      push @valid_values,$_;
    }
  }

  @valid_values = squeeze_list(@valid_values);

  my $answer;
  my $query_pattern = join("|",@valid_values);
  my $query_options;

  if (exists $options{default_value}) {    
    if ($#valid_values > 0) {
      $query_options = " (".join("/",@valid_values).")$options{terminal_mark} [$options{default_value}] ";
    }
    else {
      $query_options = "$options{terminal_mark} [$options{default_value}] ";
    }    
  }
  else {
    if ($#valid_values >= 0) {
      $query_options = " (".join("/",@valid_values).")$options{terminal_mark}";
    }
    else {
      $query_options = "$options{terminal_mark}";
    }
  }
  
 QUERY_LOOP:
  while(1) {
    if (defined $answer)
    {
      # only print the invalid choice message when the defined $answer
      # is not the default value.  Otherwise, we are handling a case
      # which has been documented (above) to be handled by the user
      # function.

      if ($options{invalid_choice} ne "") {
	print "$options{invalid_choice}: $answer.\n\n";
      }
    }
      
    print $query_string, "$query_options ";

    $answer = <STDIN>;
    chomp $answer;

    if ($answer eq "") {
      if (exists $options{default_value}) {
	$answer = $options{default_value};
      }
      else {
	redo QUERY_LOOP;
      }
    }

    if (&$query_function($answer)) {
      return($answer);
    }
  }
}

#------------------------------------------------------------------------------
# This method queries the user such that the resultant value will be matched
# against the list of valid options, provided by the implementor.  
# 
# Parameter 1 (<query_string>) is the string which will be prompted to the user
# to answer.
#
# Parameter 2 - n (<passed_valid_values>) are parameters which, among other
# things, provides a list of options which will be listed alongside the
# <query_string> as valid options. These parameters may also be keyed option 
# specifiers.  For example, "terminal_mark=?" will cause the terminal 
# character at the end of <query_string> to be '?'.  The following list of 
# options exists:
#  terminal_mark=<option> (default value: ':')
#    - changes the character(s) which succeed the <query_string>.  
#  
#  invalid_choice=<option> (default value: "Invalid choice")
#    - changes the string which gets displayed when the user inputs an invalid
#      option.
#       
#  default=<option> (default value: none)
#    - changes the default behavior of hitting return without inputing any 
#      data.  If a user hits return, with "default=foo", the resulting return
#      will be interpreted as if the user had typed foo and then hit return.
#
#  case_sensitive=[yes|no] (default value: no)
#    - specifies whether or not matches are case sensitive.
#------------------------------------------------------------------------------
sub user {
  my ($query_string, @passed_valid_values) = @_;
  
  my %options = (
		 terminal_mark => ":",
		 invalid_choice => "Invalid choice"
		);
  my @valid_values;

  foreach (@passed_valid_values) {
    if (/default\s*=\s*(.*)/si) {
      (not exists $options{default_value}) || die "More than one default value in query_user.";
      $options{default_value} = $1;
      push @valid_values,$1;
    }
    elsif (/invalid_choice\s*=\s*(.*)/si) {
      $options{invalid_choice}=$1;
    }
    elsif (/terminal_mark\s*=\s*(.*)/si) {
      $options{terminal_mark}=$1;
    }
    elsif (/case_sensitive\s*=\s*(.*)/si) {
      $options{case_sensitive} = $1;
    }
    else {
      push @valid_values,$_;
    }
  }

  @valid_values = squeeze_list(@valid_values);
  my $answer;
  my $query_pattern = join("|",@valid_values);
  my $query_options;

  if (exists $options{default_value}) {    
    if ($#valid_values > 0) {
      $query_options = " (".join("/",@valid_values).")$options{terminal_mark} [$options{default_value}] ";
    }
    else {
      $query_options = "$options{terminal_mark} [$options{default_value}] ";
    }    
  }
  else {
    if ($#valid_values >= 0) {
      $query_options = " (".join("/",@valid_values).")$options{terminal_mark}";
    }
    else {
      $query_options = "$options{terminal_mark}";
    }
  }
  
 QUERY_LOOP:
  while(1) {
    print $query_string, "$query_options ";

    $answer = <STDIN>;
    
    chomp $answer;
    if ($answer eq "") {
      if (exists $options{default_value}) {
	$answer = $options{default_value};
      }
    }
   
    if (INSTALL::conditional::boolean_value($options{case_sensitive}) &&
	   ($answer =~ /^($query_pattern)$/))
    {
      return($answer);
    }
    elsif ($answer =~ /^($query_pattern)$/si) {
      return(lc($answer))
    }
    else {
      print "Invalid choice.\n";
    }
  }
}

#------------------------------------------------------------------------------
# This method queries the user such that the resultant value will be matched
# against the list of valid options, provided by the implementor.  This method
# is similar to the user method with the exception that the values are listed
# in a menu format.  In this way, each option is given a unique index number
# which specifies acceptance of that option.  Thus, the actual option can
# be a multi-word set of text.
# 
# Parameter 1 (<menu_title>) is a parameter which provides a general 
# description of what the meaning of the menu is.
#
# Parameter 2 (<menu_text>) is a parameter which provides a description of what
# all of the options are trying to allow the user to answer.
# 
# Parameter 3 (<query_string>) is a parameter which will be displayed as a text
# query string, succeeding the list of options.  It should contain the query 
# which would be made regarding which of the choices the user would like to 
# select.
#
# Parameter 4 - n (<passed_valid_values>) are parameters which, among other
# things, provides a list of options which will be listed as valid options. 
# These parameters may also be keyed option specifiers.  For example, 
# "terminal_mark=?" will cause the terminal character at the end of 
# <query_string> to be '?'.  The following list of options exists:
#  terminal_mark=<option> (default value: ':')
#    - changes the character(s) which succeed the <query_string>.  
#  
#  invalid_choice=<option> (default value: "Invalid choice")
#    - changes the string which gets displayed when the user inputs an invalid
#      option.
#       
#  default=<option> (default value: none)
#    - changes the default behavior of hitting return without inputing any 
#      data.  If a user hits return, with "default=foo", the resulting return
#      will be interpreted as if the user had typed foo and then hit return.
#------------------------------------------------------------------------------
sub user_menu {
  my ($menu_title,$menu_text,$query_string,@passed_menu_options) = @_;
  my @menu_options;
  my %options = 
    (
     terminal_mark => ":",
     invalid_choice => "Invalid choice"
    );

  my $count = 1;
  foreach (@passed_menu_options) {
    if (/default\s*=\s*(.*)/si) {
      (not exists $options{default_value}) || die "More than one default value in query_user.";

      my $default = $1;

      # If the default contains a non-digit, it is referring to making a default item.
      if ($default =~ /\D/) {
	$options{default_value} = $count;
	push @menu_options,$default;
	$count++;
      }
      else {
	$options{default_value} = $default;
      }
    }
    elsif (/invalid_choice\s*=\s*(.*)/si) {
      $options{invalid_choice}=$1;
    }
    elsif (/terminal_mark\s*=\s*(.*)/si) {
      $options{terminal_mark}=$1;
    }
    else {
      push @menu_options,$_;
      $count++;
    }
  }

  my $menu_width = 79;
  my $header = " ".repeat_text("-",$menu_width - 2)."\n";
  my $footer = $header;
  my $separator = "|".repeat_text("-",$menu_width - 2)."|\n";
  my $empty = "|".repeat_text(" ",$menu_width - 2)."|\n";

  if (exists $options{default_value}) {
    if (($options{default_value} !~ /[0-9]+/) || 
	($options{default_value} <= 0) ||
	($options{default_value} > ($#menu_options + 1))) 
    {
      die "Invalid default value.";
    }
  }
  
  $query_string .= $options{terminal_mark};
  $query_string .= " [$options{default_value}]" if (exists $options{default_value});

  while(1) {
    system("clear");

    print("\n",$header,
	  justify_text($menu_title,$menu_width,"| ", " |"),
	  $separator,
	  justify_text($menu_text,$menu_width,"| ", " |"));
    
    print $empty if ($#menu_options >= 0);
    
    my $i = 0;
    foreach (@menu_options) {
      my $index = $i + 1;
      print justify_text("  $index. $menu_options[$i]",$menu_width,"| ", " |");
      $i = $index;
    }
    print $empty, $footer,"\n";

    if ($#_ <= 2) {
      last;
    }
    else {
      print "$query_string ";
      my $answer = <STDIN>;
      
      chomp $answer;

      if ((exists $options{default_value}) && ($answer eq "")) {
	return($options{default_value});
      }
      elsif (($answer =~ /^[0-9]+$/) && ($answer <= $i)) {
	return($answer);
      }
      else {
	my $enter_choice = "Press enter to select again ...";
	if (length("$options{invalid_choice}: $enter_choice") < 80) {
	  print "$options{invalid_choice}: $enter_choice";
	}
	else {
	  print "$options{invalid_choice}.\n$enter_choice";
	}
	<STDIN>;
      }
    }
  }
}

1;
