package INSTALL::webserver;

use POSIX;
use INSTALL::os;
use INSTALL::common;
use INSTALL::conditional;
use INSTALL::global;
use INSTALL::parameter;
use INSTALL::query;
use INSTALL::ini;
use INSTALL::apache;

my $debug = 0;
my $CONST_Package = "INSTALL::webserver";

my $pGlobals = \%INSTALL::global::variables;
my $pParameters = \%INSTALL::parameter::variables;
my $pOS = \%INSTALL::os::Details;

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Internal functions
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#------------------------------------------------------------------------------
# All of the error reporting mechanisms take of a list of strings, and format
# them such that they fall on different lines.
#------------------------------------------------------------------------------
sub error {
  my $result = INSTALL::common::generate_error(@_);
  print STDERR "Package location: $CONST_Package\n";
  return($result);
}

sub trace
{
  if ($debug)
  {
    my $result = INSTALL::common::generate_report("trace: ",@_);
    print STDERR "Package location: $CONST_Package\n";
    return($result);
  }
}

sub note
{
  my $result = INSTALL::common::generate_note(@_);
  return($result);
}

sub report 
{
  my $result = INSTALL::common::generate_report("",@_);
  return($result);
}

#------------------------------------------------------------------------------
# This method takes a single configuration file and a webserver type, and 
# collects / builds information about that server, and stores it in the 
# webserver cache.
#==============================================================================
# Implementors note:
# (1) This method calls into the respective perl module for the webserver in 
#     question (ie. it keys off of type).  It is assumed that this module will
#     contain a method named cache_server which takes one parameter, that being
#     a reference to a hash.  This hash is guaranteed to contain conf and type
#     as parameter keys.  The result is expected to be non-undef on success 
#     and undef otherwise.  
#------------------------------------------------------------------------------
sub cache_server
{
  ($#_ == 1) || die "Invalid number of arguments to add_server.";
  
  my ($config_file, $type) = @_;
  my $cache_file = "$pGlobals->{install_dir}/.webserver-cache";
      
  my %server_stats = 
    (
     conf => $config_file,
     type => $type, 
    );

  my $result;
  my $pStats = \%server_stats;

  # First, do a merge of information (that is, retrieve any information 
  # accrued thus far on the specified $config_file).  
    
  my $pOldIni = INSTALL::ini::open($cache_file);
  if (!$pOldIni)
  {
    return(error("Could not open $cache_file."));
  }
  
  my $pOldSection = INSTALL::ini::openSection($pOldIni, $server_stats{conf});
  if ($pOldSection)
  {
    # Precedence toward new information.
    %server_stats = (%{$pOldSection},%server_stats);
  }


  for($type) {
    /apache/i &&
      do {
	$result = INSTALL::apache::cache_server($pStats);
	last;
      };
    
    error("Unsupported webserver used in cache_server.");
    exit(1);
  }
  
  if ($result) {    
    my $pIni = INSTALL::ini::open($cache_file);
    if (!$pIni)
    {
      return(error("Could not open $cache_file."));
    }
    
    my $pSection = INSTALL::ini::insertSection($pIni, $server_stats{conf});
    if (!$pSection)
    {
      return(error("Could not add section $server_stats{conf} to $cache_file."));
    }

    foreach(keys %server_stats) {
      $pSection->{$_} = $server_stats{$_};
    }
    
    return(INSTALL::ini::save($pIni, $cache_file));
  }
  else {
    return($result);
  }
}


#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# External functions
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#------------------------------------------------------------------------------
# This method returns a list of the known servers.  Known here, means the list
# of servers contained in the webserver cache.
#------------------------------------------------------------------------------
sub list_servers 
{
  ($#_ < 0) || die "Invalid number of arguments to list_servers.";

  my $cache_file = "$pGlobals->{install_dir}/.webserver-cache";

  my $pIni = INSTALL::ini::open($cache_file);
  if (!$pIni)
  {
    return(error("Could not open $cache_file."));
  }
  
  return(INSTALL::ini::getSectionNames($pIni));
}


#------------------------------------------------------------------------------
# This method returns a hash containing the information about the specified
# server configuration file.  If that server conf file was not found, undef
# will be returned.
#
# Note: It should be noted that a copy of the hash is being returned, not 
# the original.  Thus, there is no way to alter the original information, 
# indirectly.
#------------------------------------------------------------------------------       
sub get_cached_info 
{
  ($#_ >= 0) || die "Invalid number of arguments to list_servers.";

  my ($server_conf, @options_list) = @_;
  my $retried = 0;
  my %options;

  if ($#_ >= 1) {
    foreach (@options_list) {
      if (/\s*([^\s=]*)\s*=\s*(.*)/) {
	my $key = $1;
	my $value = $2;
	
	if ($value =~ /\"([^\"]*|\\\"*)\"/) {
	  $options{$key} = $1;
	}
	else {
	  $options{$key} = $value;
	}
      }
      else {
	# Unnamed arguments, use incremental naming scheme.
	$options{$i++} = $_;
      }
    }  

    foreach(keys %options) {
      if ($_ eq "recache") {
	if (INSTALL::conditional::boolean_value($options{$_})) {
	  recache_server($server_conf);
	}
      }
      else {
	if (/^[0-9]+$/) {
	  note("Unrecognized option passed to get_cached_info = $options{$_}");
	}
	else {
	  note("Unrecognized option passed to get_cached_info = $_");
	}
      }
    }
  }

  my $cache_file = "$pGlobals->{install_dir}/.webserver-cache";
  my $pIni = INSTALL::ini::open($cache_file);

  if (!$pIni)
  {
    return(error("Could not open $cache_file."));
  }
  
  my $pSection = INSTALL::ini::openSection($pIni, $server_conf);
  if (not $pSection)
  {
    if (recache_server($server_conf)) {
      # Recurse.
      return(get_cached_info($server_conf));
    }
    else {
      return(error("Could not find section $server_conf in $cache_file."));
    }
  }

  return(%{$pSection});
}


#------------------------------------------------------------------------------
# Using the specified directory as a possible configuration directory, try
# to locate and cache any and all webservers located in this directory (or
# immediate subdirectories, as the rules apply).  This method will return
# 1 if at least one webserver was found, otherwise undef is returned.
#==============================================================================
# Implementors note:
# (1) It is assumed that each webserver module (ie. apache, netscape, etc). 
#     will have a locate_server method which will return two items:
#       (1) A configuration file
#       (2) The type of server (e.g., apache).
#     If the result is of smaller size, it is assumed that a failure occurred.
#------------------------------------------------------------------------------
sub locate_server
{
  ($#_ == 0) || die "Invalid number of arguments.";

  my ($conf_path) = @_;
  my @servers;

  push @servers,INSTALL::apache::locate_server($conf_path);
  ####### Add more webservers as they come along. #######


  if ($#servers > 0) {
    for(my $i=0; $i < $#servers; $i+=2) {
      cache_server($servers[i*2],$servers[$i*2 + 1]);
    }

    return(1);
  }
  else {
    return(undef);
  }
}


#------------------------------------------------------------------------------
# Recursively traverse all directories starting at the top-level directory
# specified by <search_path>, attempting to find all webservers contained
# under that directory.  Upon successfully establishing at least one webserver,
# 1 is returned.  Otherwise, undef is returned.  The optional second argument
# specifies the maximum depth that the call may traverse to.
#------------------------------------------------------------------------------
sub locate_servers
{
  # Search only local filesystems ... (if we can tell the difference).
  my @mount = split(/\n/,`mount`);
  my %disabled_directories = ("/proc" => 1,"/dev" => 1);

  foreach(@mount) {
    my $line = $_;
    
    # OS specific.
    for($pOS->{os}) {
      /linux/i &&
	do {
	  if ($line =~ /on\s+(\S+)\s+type\s+nfs/) {
	    $disabled_directories{$1} = 1;
	  }
	  last;
	};

      /sunos/i &&
	do {
	  if ($line =~ m@(\S+).*/remote @) {
	    $disabled_directories{$1} = 1;	    
	  }
	  last;
	};

      /aix/i &&
	do {
	  if ($line =~ /(\S+)\s+nfs/) {
	    $disabled_directories{$1} = 1;
	  }
	  last;
	};
    }
  }

  sub internal_locate {
    my ($search_path,$depth) = @_;
    my $result;

    my @subdirs;

    if ($depth <= 0) {
      return(undef);
    }

    unless(opendir(DIR,"$search_path")) {
      return(error("Unable to open $search_path in locate_servers."));
    }

    if ($search_path eq '/') {
      $search_path = "";
    }
  
    foreach(readdir(DIR)) {
      my $path = "$search_path/$_";
      my $file = $_;

      if ((not -l $path) && 
	  (-d $path) &&
	  ($file ne ".") && 
	  ($file ne "..") &&
	  (not $disabled_directories{$path})) 
      {
	push @subdirs, $path;
      }
    }
    
    close(DIR);
    foreach(@subdirs) {
      if (locate_server($_)) {
	$result = 1;
      }
      
      if (internal_locate($_,$depth-1)) {
	$result = 1;
      }
    }    

    return($result);
  }

  ($#_ == 0) ||  ($#_ == 1) || die "Invalid number of arguments.";

  my ($search_path, $depth) = @_;
  my $result;

  if (not $depth) {
    $depth = 1000000;
  }

  if (not length($search_path)) {
    $search_path = `pwd`;
  }

  $result = locate_server($search_path);
  $result = internal_locate($search_path,$depth);
  
  return($result);
}

#------------------------------------------------------------------------------
# This method allows a user to attempt to recache a specific webserver.  To
# do this, pass in the associated configuration file of that webserver.  Upon
# successful recaching, a value which will evaluate to true is returned.  
# Otherwise, undef is returned.  It should be noted that this method will error
# out immediately if the specified webserver configuration file does not exist.
#------------------------------------------------------------------------------
sub recache_server
{
  ($#_ == 0) || die "Invalid number of arguments.";

  my ($server_conf) = @_;

  if (-f $server_conf) {
    my $conf_root = $server_conf;
    $conf_root =~ s@^(.*)/[^/]*$@$1@;
    
    return(locate_server($conf_root));
  }
  else {
    return(undef);
  }
}

#------------------------------------------------------------------------------
# This is an interactive (or non_interactive, if appropriate parameters are
# provided via INSTALL::parameter) method which will attempt to discover the
# location of all of the installed webservers on the box.  In the interactive
# variant, the user will be queried for how aggressive the search will be).  In
# the non_interactive variant, the specified path (specified via 
# $INSTALL::parameter::variables{webserver_path}) will be searched for valid
# servers.  If no server are found, undef will be returned.  Otherwise, 1 is
# returned.
#------------------------------------------------------------------------------
sub build_server_cache
{
  my @scandirs = @_;
  my $result = undef;

  if (exists $pParameters->{webserver_path}) {
    if (locate_server($pParameters->{webserver_path})) {
      $result = 1;
    }
  }
  elsif (exists $pParameters->{webserver_conf}) {
    my $parent_dir = $pParameters->{webserver_conf};
    $parent_dir =~ s@^(.*)/[^/]*$@$1@;

    if (locate_server($parent_dir)) {
      $result = 1;
    }
  }
  elsif ($#scandirs >= 0) {
    foreach(@scandirs) {
      if (locate_servers($_)) {
	$result = 1;
      }
    }
  }
  else {
    my $type = INSTALL::query::user_menu
      ("WEBSERVER SEARCH",
       "For Chili!Soft ASP to simplify the installation process, it can search ".
       "your system for webservers to attach itself to.  However, this ability ".
       "comes with a cost in search time.  To give the user some ".
       "flexibility, several options have been created to manage the amount of ".
       "time required for this search.",
       "Which type of search would you like to perform",
       "terminal_mark=?",
       "Perform an exhaustive search of the entire system (slow).",
       "Perform a search of the /usr, /etc, and /var directories (moderate).",
       "Perform a search of common locations for webservers (fast).",
       "Don't perform any search.  Instead, require user specification.",
       "default=1");

    if ($type == 4) {
      $result = 1;
    }
    else {
      print STDERR ("Searching for webservers ... ");
      
      my $pid = fork;
      
      if (not $pid) {
	if ($type == 1) {
	  $result = locate_servers("/");
	}
	elsif ($type == 2) {
	  $result = locate_servers("/usr");
	  $result |= locate_servers("/etc");
	  $result |= locate_servers("/var");
	}
	elsif ($type == 3) {
	  $result = locate_servers("/etc",1);
	  $result |= locate_servers("/opt",1);
	  $result |= locate_servers("/usr/local",1);
	  $result |= locate_servers("/usr",1);
	  
	  # /var and /var/lib are likely candidates for roots of webservers (Apache).
	  # on Slackware, et al.
	  $result |= locate_servers("/var",2);
	}
	
	exit($result ? 1 : 0);
      }
      else {
	INSTALL::common::spinner_wait($pid);
	print "\n";
	
	if ($?) {
	  $result = 1;
	}
      }
    }
  }

  # Confirm that all of the conf files listed in the webserver cache are valid.
  # If not create a remove list, then remove them.
  my @delete_list;
  foreach(list_servers) {
    if (not -f $_) {
      push @delete_list,$_;
    }
  }

  if ($#delete_list >= 0) {
    my $cache_file = "$pGlobals->{install_dir}/.webserver-cache";
    my $pIni = INSTALL::ini::open($cache_file);
    
    if (!$pIni)
    {
      return(error("Could not open $cache_file -- build_server_cache."));
    }

    foreach(@delete_list) {
      INSTALL::ini::deleteSection($pIni,$_);
    }
    
    INSTALL::ini::save($pIni,$cache_file);
  }

  return($result);
}

#------------------------------------------------------------------------------
# This method is the main interface for installing an unlisted server.  It will
# call into the appropriate webserver pm, to configure the webserver in a 
# webserver dependant fashion.
#==============================================================================
# Implementors note:
# (1) Each webserver pm is assumed to have a configure_unlisted method which 
#     will query the user for all of the pertinent information about the 
#     server, and which will return a hash containing that information.  If
#     the resultant hash does not contain a conf member (which specifies the
#     servers document root directory), it is assumed that the configuration
#     failed.  It should be ****NOTED**** that the actual configuration of the
#     server does not happen here, but instead in configure.
#------------------------------------------------------------------------------
sub configure_unlisted
{
  my @webservers = ("Apache");
  # List more webservers here as they become appropriate
  
  my $cache_file = "$pGlobals->{install_dir}/.webserver-cache";
  my $type = INSTALL::query::user_menu
    ("UNLISTED SERVER CONFIGURATION",
     "Listed below are the types of servers which this version of Chili!Soft ".
     "ASP will install itself to.",
     "Which webserver type do you wish to install",
     "terminal_mark=?",
     @webservers,
     "Cancel unlisted server configuration."
    );
  
  my $result = undef;

  if ($type <= ($#webservers + 1)) {
    my %server_stats;

    for($webservers[$type - 1]) {
      /apache/i && 
	do {
	  %server_stats = INSTALL::apache::configure_unlisted;
	  last;
	};
      
      error("Unsupported webserver used in configure_unlisted.");
      exit(1);
    }

    if (exists $server_stats{conf}) {
      $result = 1;

      my $pIni = INSTALL::ini::open($cache_file);
      if (!$pIni)
      {
	return(error("Could not open $cache_file - configure_unlisted."));
      }
      
      my $pSection = INSTALL::ini::insertSection($pIni, $server_stats{conf});
      if (!$pSection)
      {
	return(error("Could not add section $server_stats{conf} to $cache_file - configure_unlisted."));
      }
      
      foreach(keys %server_stats) {
	$pSection->{$_} = $server_stats{$_};
      }
      
      $result = INSTALL::ini::save($pIni, $cache_file);
    }    
  }

  return($result);
}

#------------------------------------------------------------------------------
# This method attempts to discover a link between the specified webserver 
# configuration file, and any currently installed Chili!ASP engine cluster.
# If no such linkage exists, undef is returned.  Otherwise, the information
# regarding the specified server is returned in a hash.
#==============================================================================
# Implementor's note:
# (1) This depends upon the installed_db cache containing engine entrees 
#     specifying a key of 'engine_name'.
# (2) Also required is the 'webserver_conf' key.
#------------------------------------------------------------------------------
sub discover_linked_server 
{
  ($#_ == 0) || die "Invalid number of arguments.";
  
  my ($conf) = @_;
  
  my $pIni = INSTALL::ini::open($pGlobals->{installed_db});
  if (!$pIni)
  {
    return(error("Could not open $pGlobals->{installed_db} - discover_linked_server."));
  }
  
  my @section_list = INSTALL::ini::getSectionNames($pIni);
  my %server_details;
  
  foreach(@section_list) {
    my $pSection = INSTALL::ini::openSection($pIni,$_);
    if ($pSection->{webserver_conf} eq $conf) {
      %server_details = %{$pSection};
    }
  }
  
  return(%server_details);    
}


#------------------------------------------------------------------------------
# This method queries the user for which server they want to install to (of 
# those listed in the webserver cache).  If there are none in the cache, and 
# they choose unlisted, the unlisted server will be checked for existance 
# prior to installation.
#==============================================================================
# Implementors note:
# (1) Each webserver pm is assumed to have a configure method which will take
#     a hash as a parameter and use the information contained in that hash
#     to configure a server.  This method should return undef upon failure, 
#     non-undef on success.
#------------------------------------------------------------------------------
sub configure
{
  sub interactive_configure
  {
  RETRY_CONFIGURE:
    my @webservers = list_servers;
    my $webservers_count = $#webservers + 1;
    my @webserver_options = ();
    
    foreach(@webservers) {
      my %cached_info = get_cached_info($_);
      my $server_text = ucfirst($cached_info{type});
      
      if (INSTALL::conditional::boolean_value($cached_info{https})) {
	$server_text .= " Secure Server";
      }
      
      $server_text .= " ($_)";
      
      my %server_info = discover_linked_server($_);
      if (keys %server_info) {
	$server_text .= " <<$server_info{engine_name}>>";
      }
      
      $server_text .= ".";
      
      push @webserver_options,$server_text;
    }
    
    my $index = INSTALL::query::user_menu
      ("CONFIGURING YOUR WEBSERVER",
       "Setup will automatically configure your webserver.  Choose which webserver you would like to use below.",
       "Which webserver would you like to configure", 
       "terminal_mark=?", 
       @webserver_options,
       "Add an unlisted webserver.",
       "Refresh the webserver list.",
       "Do not configure a webserver.",
      );
    
    print "\n";
    if ($index == ($webservers_count + 1)) {
      configure_unlisted;
      goto RETRY_CONFIGURE;
    }
    elsif ($index == ($webservers_count + 2)) {
      build_server_cache;
      goto RETRY_CONFIGURE;
    }
    elsif ($index == ($webservers_count + 3)) {
      return;
    }
    else {
      $index--;
      
      my %server_link = discover_linked_server($webservers[$index]);
      
      if (keys %server_link) {
	note("You cannot install to a webserver which has already been linked to",
	     "by another server.  The currently linked Chili!Soft ASP server is:",
	     "  $server_link{engine_name}");
	
	my $uninstall_link = INSTALL::query::user
	  ("Would you like to uninstall the currently linked server",
	   "terminal_mark=?",
	   "y",
	   "default=n");
	
	if ($uninstall_link eq "y") {
	  uninstall($server_link{engine_dir});
	}
      }
      
      my %webserver_stats = get_cached_info($webservers[$index],"recache=true");
      
      INSTALL::query::user_menu
	("WEBSERVER INFORMATION CONFIRMATION",
	 "Make certain that all of the information listed below is accurate, to the ".
	 "best of your knowledge.  If any of this information is incorrect, it can ".
	 "be corrected by answering no to the question below, and selecting 'Add ".
	 "an unlisted webserver.' on the previous menu.  You will then be allowed to ".
	 "alter certain important details about the selected webserver.  As a preliminary ".
	 "step, please write down the full path of the configuration file (listed below).".
	 "\n \n".
	 "Webserver information:\n".
	 "  Webserver main configuration file: $webserver_stats{conf}\n".
	 "  Webserver binary: $webserver_stats{binary}\n".
	 "  Webserver version: $webserver_stats{version}\n".
	 "  Webserver type: ".ucfirst($webserver_stats{type})."\n".
	 "  Webserver port: $webserver_stats{port}\n".
	 "  Webserver root: $webserver_stats{root}\n"
	);
      
      my $confirm_install = INSTALL::query::user
	("Does all of the information seem to be correct","terminal_mark=?",
	 "y","default=n");

      if ($confirm_install eq "n") {
	goto RETRY_CONFIGURE;
      }
      
      my $install_defaults;
      
      $install_defaults = INSTALL::query::user_menu
	("GENERAL INSTALLATION OPTIONS",
	 "During the next phase of the installation, there are two drastically different ".
	 "modes that the installer can run in.  The 'Default' mode will only prompt users ".
	 "for information which it can neither auto-detect / select the default settings ".
	 "for.  Choosing the 'Customize' mode will allow you to pick and choose from the ".
	 "available feature-set, as well as to set specific values regarding the overall ".
	 "configuration.  It is recommended that your first Chili!Soft ASP experience use ".
	 "the default mode, since it will allow you to get accustomed with working with ".
	 "Chili!Soft ASP with its 'stock' settings.",
	 "Which mode would you like to run in",
	 "terminal_mark=?",
	 "default=Auto-Detect / Select Default settings.",
	 "Customize settings.",
	 "Choose another webserver to install to.");
      
      if ($install_defaults == 2) {
	$pParameters->{custom_install} = "true";	
      }
      elsif ($install_defaults == 3) {
	goto RETRY_CONFIGURE;
      }

      return(%webserver_stats);
    }
  }

  my $interactive = (not INSTALL::conditional::boolean_value($pParameters->{non_interactive}));
  my  %cached_info;

  if ($interactive) {
    %cached_info = interactive_configure;
  }
  else {
    if (exists $pParameters->{webserver_path}) {
      my $quoted_webserver_path = quotemeta($pParameters->{webserver_path});
      foreach(list_servers) {
	if (m@^$quoted_webserver_path@) {
	  %cached_info = get_cached_info($_,"recache=yes");
	  last;
	}
      }
    }
    elsif (exists $pParameters->{webserver_conf}) {
      if (grep { $pParameters->{webserver_conf} eq $_ } list_servers) {	
	%cached_info = get_cached_info($pParameters->{webserver_conf},"recache=yes");
      }
      else {
	build_server_cache;
	
	if (grep { $pParameters->{webserver_conf} eq $_ } list_servers) {	
	  %cached_info = get_cached_info($pParameters->{webserver_conf});
	}	  
      }
    }
    else {
      die("Either the variable webserver_path or webserver_conf must be set equal to\n",
	  "something, in non_interactive mode.");
    }
      

    if (not keys %cached_info) {
      return(error("Unable to find a webserver in the specified webserver_path / webserver_conf:",
		   "  $pParameters->{webserver_path} / $pParameters->{webserver_path}"));
    }
    else {
      my %server_link = discover_linked_server($cached_info{conf});

      if (keys %server_link) {
	uninstall($server_link{engine_dir});
      }
    }
  }

  my $result = undef;
  if (keys %cached_info) {
    for($cached_info{type}) {
      /apache/i &&
	do {
	  $result = INSTALL::apache::configure(%cached_info);
	  last;
	};
      
      error("Unsupported webserver used in configure.");
      exit(1);
    }
  }

  # Don't persist the custom_install flag.
  delete $pParameters->{custom_install};

  return($result);
}

#------------------------------------------------------------------------------
# This method lists all used servers.  Each servers information is placed into
# the returned list as a reference to a hash.
#==============================================================================
#------------------------------------------------------------------------------
sub list_used_servers 
{
  my $pIni = INSTALL::ini::open($pGlobals->{installed_db});
  if (!$pIni)
  {
    return(error("Could not open $pGlobals->{installed_db} - list_used_servers."));
  }
  
  my @section_list = INSTALL::ini::getSectionNames($pIni);
  my @result;
  
  foreach(@section_list) {
    my $pSection = INSTALL::ini::openSection($pIni,$_);
    my %server_details = %{$pSection};
    
    if (keys %server_details) {
      push @result,\%server_details;
    }
  }
  
  return(@result);
}


#------------------------------------------------------------------------------
# This method queries the user for which server they want to uninstall (of 
# those listed in the installed_db).  If there are none in the cache, this
# method will merely return success (ie. non-undef).  The optional arguments
# are the specific list of engine directories to uninstall.
#
# Note: This method is a general means of also removing the asp-engine which 
# is serving that webserver.
#==============================================================================
# Implementors note:
# (1) Each webserver pm is assumed to have an uninstall method which will take
#     a hash as a parameter and use the information contained in that hash
#     to uninstall that server.  This method should return undef upon failure, 
#     non-undef on success.  It should be noted that this method is responsible
#     for also invoking caspeng::uninstall, with the appropriate information.
#------------------------------------------------------------------------------
sub uninstall
{
  sub interactive_uninstall
  {
    my @params = @_;
    my @servers = list_used_servers;
    my @options;
    my $chosen_default;
    
    if ($#servers >= 0) {
      foreach(@servers) {
	my $pServerInfo = $_;
	my $server_string = 
	  ("$pServerInfo->{engine_name} associated with:\n".
#	   "                               Type: ".ucfirst($pServerInfo->{webserver_type})."\n".
	   "       Webserver configuration file: $pServerInfo->{webserver_conf}\n".
	   "              Webserver port number: $pServerInfo->{webserver_port}\n"
	  );

	if ((not $chosen_default) &&
	    ($#params == 0) && 
	    (grep { $pServerInfo->{engine_dir} eq $_ } @params))
	{
	  $chosen_default = 1;
	  $server_string = "default=$server_string";
	}
	
	push @options,$server_string;
      }
    }
    else {
      INSTALL::query::user_menu
	("ENGINE UNINSTALL",
	 "No engines are available for uninstall.");
      return;
    }

    my $max_options =$#options + 1;

    if (not $chosen_default) {
      # If the default wasn't preselected, choose a default of "Cancel uninstall."
      push @options,("default=".($max_options+2));
    }

  RETRY_UNINSTALL_QUERY:    
    my $server_no = INSTALL::query::user_menu
      ("ENGINE UNINSTALL",
       "By uninstalling a Chili!Soft ASP engine, you will be removing the link ".
       "between the associated webserver and the Chili!Soft ASP engine.  It should ".
       "be noted that any data contained under the specified engine directory ".
       "will be destroyed along with that engine.",
       "Select the server you want uninstalled",
       @options,
       "Uninstall all engines.",
       "Cancel uninstall.");

    my @uninstall_list;
    if ($server_no <= $max_options) {
      report("Warning: You are about to uninstall the engine $servers[$server_no - 1]->{engine_name}.");

      my $answer = INSTALL::query::user
	("Do you wish to continue","terminal_mark=?","y","default=n");

      if ($answer eq "n") {
	goto RETRY_UNINSTALL_QUERY;
      }

      push @uninstall_list,$servers[$server_no - 1];
    }
    elsif ($server_no == ($max_options + 1)) {
      report("Warning: You are about to uninstall all engine associations.");
      my $answer = INSTALL::query::user
	("Do you wish to continue","terminal_mark=?","y","default=n");

      if ($answer eq "n") {
	goto RETRY_UNINSTALL_QUERY;
      }

      @uninstall_list = @servers;
    }
    else {
      return;
    }

    return(@uninstall_list);
  }

  my @params = @_;
  my $interactive = (not INSTALL::conditional::boolean_value($pParameters->{non_interactive}));
  my @uninstall_list;

 RETRY_UNINSTALL:
  if ($interactive) {
    @uninstall_list = interactive_uninstall(@params);
  }
  elsif ($#params >= 0) {
    my @servers = list_used_servers;
    
    foreach(@servers) {
      my $pServer = $_;

      if (grep { $_ eq $pServer->{engine_dir}} @params) {
	push @uninstall_list, $_;
      }
    }
  }	
  else {
    return(undef);
  }
	
  my $result = 1;
  foreach(@uninstall_list) {
    my $pServerInfo = $_;

    for($pServerInfo->{webserver_type}) {
      /apache/i && 
	do {
	  unless (INSTALL::apache::uninstall(%{$pServerInfo})) {
	    error("Failed to uninstall $pServerInfo->{engine_name}.");
	    print STDERR "Press enter to continue ...";
	    <STDIN>;
	    goto RETRY_UNINSTALL;
	  }
	  last;
	};

      return(error("Unsupported webserver encountered."));
    }

    if ($interactive) {
      if ($result) {
	report("Status: The uninstall was successful.");
      }
      else {
	report("Status: The uninstall failed.");
      }
    }
  }

  return($result && ($#uninstall_list >= 0));
}

1;
